"""
Pattern discovery involves the identification of musical patterns (i.e. short
fragments or melodic ideas that repeat at least twice) both from audio and
symbolic representations.  The metrics used to evaluate pattern discovery
systems attempt to quantify the ability of the algorithm to not only determine
the present patterns in a piece, but also to find all of their occurrences.

Based on the methods described here:
    T. Collins. MIREX task: Discovery of repeated themes & sections.
    http://www.music-ir.org/mirex/wiki/2013:Discovery_of_Repeated_Themes_&_Sections,
    2013.

Conventions
-----------

The input format can be automatically generated by calling
:func:`mir_eval.io.load_patterns`.  This format is a list of a list of
tuples.  The first list collections patterns, each of which is a list of
occurences, and each occurrence is a list of MIDI onset tuples of
``(onset_time, mid_note)``

A pattern is a list of occurrences. The first occurrence must be the prototype
of that pattern (i.e. the most representative of all the occurrences).  An
occurrence is a list of tuples containing the onset time and the midi note
number.

Metrics
-------

* :func:`mir_eval.pattern.standard_FPR`: Strict metric in order to find the
  possibly transposed patterns of exact length. This is the only metric that
  considers transposed patterns.
* :func:`mir_eval.pattern.establishment_FPR`: Evaluates the amount of patterns
  that were successfully identified by the estimated results, no matter how
  many occurrences they found.  In other words, this metric captures how the
  algorithm successfully *established* that a pattern repeated at least twice,
  and this pattern is also found in the reference annotation.
* :func:`mir_eval.pattern.occurrence_FPR`: Evaluation of how well an estimation
  can effectively identify all the occurrences of the found patterns,
  independently of how many patterns have been discovered. This metric has a
  threshold parameter that indicates how similar two occurrences must be in
  order to be considered equal.  In MIREX, this evaluation is run twice, with
  thresholds .75 and .5.
* :func:`mir_eval.pattern.three_layer_FPR`: Aims to evaluate the general
  similarity between the reference and the estimations, combining both the
  establishment of patterns and the retrieval of its occurrences in a single F1
  score.
* :func:`mir_eval.pattern.first_n_three_layer_P`: Computes the three-layer
  precision for the first N patterns only in order to measure the ability of
  the algorithm to sort the identified patterns based on their relevance.
* :func:`mir_eval.pattern.first_n_target_proportion_R`: Computes the target
  proportion recall for the first N patterns only in order to measure the
  ability of the algorithm to sort the identified patterns based on their
  relevance.
"""


import numpy as np
from . import util
import warnings
import collections


def _n_onset_midi(patterns):
    """Computes the number of onset_midi objects in a pattern

    Parameters
    ----------
    patterns :
        A list of patterns using the format returned by
        :func:`mir_eval.io.load_patterns()`

    Returns
    -------
    n_onsets : int
        Number of onsets within the pattern.

    """
    return len([o_m for pat in patterns for occ in pat for o_m in occ])


def validate(reference_patterns, estimated_patterns):
    """Checks that the input annotations to a metric look like valid pattern
    lists, and throws helpful errors if not.

    Parameters
    ----------
    reference_patterns : list
        The reference patterns using the format returned by
        :func:`mir_eval.io.load_patterns()`
    estimated_patterns : list
        The estimated patterns in the same format

    Returns
    -------

    """
    # Warn if pattern lists are empty
    if _n_onset_midi(reference_patterns) == 0:
        warnings.warn('Reference patterns are empty.')
    if _n_onset_midi(estimated_patterns) == 0:
        warnings.warn('Estimated patterns are empty.')
    for patterns in [reference_patterns, estimated_patterns]:
        for pattern in patterns:
            if len(pattern) <= 0:
                raise ValueError("Each pattern must contain at least one "
                                 "occurrence.")
            for occurrence in pattern:
                for onset_midi in occurrence:
                    if len(onset_midi) != 2:
                        raise ValueError("The (onset, midi) tuple must "
                                         "contain exactly 2 elements.")


def _occurrence_intersection(occ_P, occ_Q):
    """Computes the intersection between two occurrences.

    Parameters
    ----------
    occ_P : list of tuples
        (onset, midi) pairs representing the reference occurrence.
    occ_Q : list
        second list of (onset, midi) tuples

    Returns
    -------
    S : set
        Set of the intersection between occ_P and occ_Q.

    """
    set_P = set([tuple(onset_midi) for onset_midi in occ_P])
    set_Q = set([tuple(onset_midi) for onset_midi in occ_Q])
    return set_P & set_Q    # Return the intersection


def _compute_score_matrix(P, Q, similarity_metric="cardinality_score"):
    """Computes the score matrix between the patterns P and Q.

    Parameters
    ----------
    P : list
        Pattern containing a list of occurrences.
    Q : list
        Pattern containing a list of occurrences.
    similarity_metric : str
        A string representing the metric to be used
        when computing the similarity matrix. Accepted values:
        - "cardinality_score":
            Count of the intersection between occurrences.
        (Default value = "cardinality_score")

    Returns
    -------
    sm : np.array
        The score matrix between P and Q using the similarity_metric.

    """
    sm = np.zeros((len(P), len(Q)))     # The score matrix
    for iP, occ_P in enumerate(P):
        for iQ, occ_Q in enumerate(Q):
            if similarity_metric == "cardinality_score":
                denom = float(np.max([len(occ_P), len(occ_Q)]))
                # Compute the score
                sm[iP, iQ] = len(_occurrence_intersection(occ_P, occ_Q)) / \
                    denom
            # TODO: More scores: 'normalised matching socre'
            else:
                raise ValueError("The similarity metric (%s) can only be: "
                                 "'cardinality_score'.")
    return sm


def standard_FPR(reference_patterns, estimated_patterns, tol=1e-5):
    """Standard F1 Score, Precision and Recall.

    This metric checks if the prototype patterns of the reference match
    possible translated patterns in the prototype patterns of the estimations.
    Since the sizes of these prototypes must be equal, this metric is quite
    restictive and it tends to be 0 in most of 2013 MIREX results.

    Examples
    --------
    >>> ref_patterns = mir_eval.io.load_patterns("ref_pattern.txt")
    >>> est_patterns = mir_eval.io.load_patterns("est_pattern.txt")
    >>> F, P, R = mir_eval.pattern.standard_FPR(ref_patterns, est_patterns)

    Parameters
    ----------
    reference_patterns : list
        The reference patterns using the format returned by
        :func:`mir_eval.io.load_patterns()`
    estimated_patterns : list
        The estimated patterns in the same format
    tol : float
        Tolerance level when comparing reference against estimation.
        Default parameter is the one found in the original matlab code by
        Tom Collins used for MIREX 2013.
        (Default value = 1e-5)

    Returns
    -------
    f_measure : float
        The standard F1 Score
    precision : float
        The standard Precision
    recall : float
        The standard Recall

    """
    validate(reference_patterns, estimated_patterns)
    nP = len(reference_patterns)    # Number of patterns in the reference
    nQ = len(estimated_patterns)    # Number of patterns in the estimation
    k = 0                           # Number of patterns that match

    # If no patterns were provided, metric is zero
    if _n_onset_midi(reference_patterns) == 0 or \
       _n_onset_midi(estimated_patterns) == 0:
        return 0., 0., 0.

    # Find matches of the prototype patterns
    for ref_pattern in reference_patterns:
        P = np.asarray(ref_pattern[0])      # Get reference prototype
        for est_pattern in estimated_patterns:
            Q = np.asarray(est_pattern[0])  # Get estimation prototype

            if len(P) != len(Q):
                continue

            # Check transposition given a certain tolerance
            if (len(P) == len(Q) == 1 or
                    np.max(np.abs(np.diff(P - Q, axis=0))) < tol):
                k += 1
                break

    # Compute the standard measures
    precision = k / float(nQ)
    recall = k / float(nP)
    f_measure = util.f_measure(precision, recall)
    return f_measure, precision, recall


def establishment_FPR(reference_patterns, estimated_patterns,
                      similarity_metric="cardinality_score"):
    """Establishment F1 Score, Precision and Recall.

    Examples
    --------
    >>> ref_patterns = mir_eval.io.load_patterns("ref_pattern.txt")
    >>> est_patterns = mir_eval.io.load_patterns("est_pattern.txt")
    >>> F, P, R = mir_eval.pattern.establishment_FPR(ref_patterns,
    ...                                              est_patterns)


    Parameters
    ----------
    reference_patterns : list
        The reference patterns in the format returned by
        :func:`mir_eval.io.load_patterns()`

    estimated_patterns : list
        The estimated patterns in the same format

    similarity_metric : str
        A string representing the metric to be used when computing the
        similarity matrix. Accepted values:

            - "cardinality_score": Count of the intersection
              between occurrences.

        (Default value = "cardinality_score")


    Returns
    -------
    f_measure : float
        The establishment F1 Score
    precision : float
        The establishment Precision
    recall : float
        The establishment Recall

    """
    validate(reference_patterns, estimated_patterns)
    nP = len(reference_patterns)    # Number of elements in reference
    nQ = len(estimated_patterns)    # Number of elements in estimation
    S = np.zeros((nP, nQ))          # Establishment matrix

    # If no patterns were provided, metric is zero
    if _n_onset_midi(reference_patterns) == 0 or \
       _n_onset_midi(estimated_patterns) == 0:
        return 0., 0., 0.

    for iP, ref_pattern in enumerate(reference_patterns):
        for iQ, est_pattern in enumerate(estimated_patterns):
            s = _compute_score_matrix(ref_pattern, est_pattern,
                                      similarity_metric)
            S[iP, iQ] = np.max(s)

    # Compute scores
    precision = np.mean(np.max(S, axis=0))
    recall = np.mean(np.max(S, axis=1))
    f_measure = util.f_measure(precision, recall)
    return f_measure, precision, recall


def occurrence_FPR(reference_patterns, estimated_patterns, thres=.75,
                   similarity_metric="cardinality_score"):
    """Establishment F1 Score, Precision and Recall.


    Examples
    --------
    >>> ref_patterns = mir_eval.io.load_patterns("ref_pattern.txt")
    >>> est_patterns = mir_eval.io.load_patterns("est_pattern.txt")
    >>> F, P, R = mir_eval.pattern.occurrence_FPR(ref_patterns,
    ...                                           est_patterns)


    Parameters
    ----------
    reference_patterns : list
        The reference patterns in the format returned by
        :func:`mir_eval.io.load_patterns()`
    estimated_patterns : list
        The estimated patterns in the same format
    thres : float
        How similar two occcurrences must be in order to be considered
        equal
        (Default value = .75)
    similarity_metric : str
        A string representing the metric to be used
        when computing the similarity matrix. Accepted values:

            - "cardinality_score": Count of the intersection
              between occurrences.

        (Default value = "cardinality_score")


    Returns
    -------
    f_measure : float
        The establishment F1 Score
    precision : float
        The establishment Precision
    recall : float
        The establishment Recall

    """
    validate(reference_patterns, estimated_patterns)
    # Number of elements in reference
    nP = len(reference_patterns)
    # Number of elements in estimation
    nQ = len(estimated_patterns)
    # Occurrence matrix with Precision and recall in its last dimension
    O_PR = np.zeros((nP, nQ, 2))

    # Index of the values that are greater than the specified threshold
    rel_idx = np.empty((0, 2), dtype=int)

    # If no patterns were provided, metric is zero
    if _n_onset_midi(reference_patterns) == 0 or \
       _n_onset_midi(estimated_patterns) == 0:
        return 0., 0., 0.

    for iP, ref_pattern in enumerate(reference_patterns):
        for iQ, est_pattern in enumerate(estimated_patterns):
            s = _compute_score_matrix(ref_pattern, est_pattern,
                                      similarity_metric)
            if np.max(s) >= thres:
                O_PR[iP, iQ, 0] = np.mean(np.max(s, axis=0))
                O_PR[iP, iQ, 1] = np.mean(np.max(s, axis=1))
                rel_idx = np.vstack((rel_idx, [iP, iQ]))

    # Compute the scores
    if len(rel_idx) == 0:
        precision = 0
        recall = 0
    else:
        P = O_PR[:, :, 0]
        precision = np.mean(np.max(P[np.ix_(rel_idx[:, 0], rel_idx[:, 1])],
                                   axis=0))
        R = O_PR[:, :, 1]
        recall = np.mean(np.max(R[np.ix_(rel_idx[:, 0], rel_idx[:, 1])],
                                axis=1))
    f_measure = util.f_measure(precision, recall)
    return f_measure, precision, recall


def three_layer_FPR(reference_patterns, estimated_patterns):
    """Three Layer F1 Score, Precision and Recall. As described by Meridith.

    Examples
    --------
    >>> ref_patterns = mir_eval.io.load_patterns("ref_pattern.txt")
    >>> est_patterns = mir_eval.io.load_patterns("est_pattern.txt")
    >>> F, P, R = mir_eval.pattern.three_layer_FPR(ref_patterns,
    ...                                            est_patterns)

    Parameters
    ----------
    reference_patterns : list
        The reference patterns in the format returned by
        :func:`mir_eval.io.load_patterns()`
    estimated_patterns : list
        The estimated patterns in the same format

    Returns
    -------
    f_measure : float
        The three-layer F1 Score
    precision : float
        The three-layer Precision
    recall : float
        The three-layer Recall

    """
    validate(reference_patterns, estimated_patterns)

    def compute_first_layer_PR(ref_occs, est_occs):
        """Computes the first layer Precision and Recall values given the
        set of occurrences in the reference and the set of occurrences in the
        estimation.

        Parameters
        ----------
        ref_occs :

        est_occs :


        Returns
        -------

        """
        # Find the length of the intersection between reference and estimation
        s = len(_occurrence_intersection(ref_occs, est_occs))

        # Compute the first layer scores
        precision = s / float(len(ref_occs))
        recall = s / float(len(est_occs))
        return precision, recall

    def compute_second_layer_PR(ref_pattern, est_pattern):
        """Computes the second layer Precision and Recall values given the
        set of occurrences in the reference and the set of occurrences in the
        estimation.

        Parameters
        ----------
        ref_pattern :

        est_pattern :


        Returns
        -------

        """
        # Compute the first layer scores
        F_1 = compute_layer(ref_pattern, est_pattern)

        # Compute the second layer scores
        precision = np.mean(np.max(F_1, axis=0))
        recall = np.mean(np.max(F_1, axis=1))
        return precision, recall

    def compute_layer(ref_elements, est_elements, layer=1):
        """Computes the F-measure matrix for a given layer. The reference and
        estimated elements can be either patters or occurrences, depending
        on the layer.

        For layer 1, the elements must be occurrences.
        For layer 2, the elements must be patterns.

        Parameters
        ----------
        ref_elements :

        est_elements :

        layer :
             (Default value = 1)

        Returns
        -------

        """
        if layer != 1 and layer != 2:
            raise ValueError("Layer (%d) must be an integer between 1 and 2"
                             % layer)

        nP = len(ref_elements)      # Number of elements in reference
        nQ = len(est_elements)      # Number of elements in estimation
        F = np.zeros((nP, nQ))      # F-measure matrix for the given layer
        for iP in range(nP):
            for iQ in range(nQ):
                if layer == 1:
                    func = compute_first_layer_PR
                elif layer == 2:
                    func = compute_second_layer_PR

                # Compute layer scores
                precision, recall = func(ref_elements[iP], est_elements[iQ])
                F[iP, iQ] = util.f_measure(precision, recall)
        return F

    # If no patterns were provided, metric is zero
    if _n_onset_midi(reference_patterns) == 0 or \
       _n_onset_midi(estimated_patterns) == 0:
        return 0., 0., 0.

    # Compute the second layer (it includes the first layer)
    F_2 = compute_layer(reference_patterns, estimated_patterns, layer=2)

    # Compute the final scores (third layer)
    precision_3 = np.mean(np.max(F_2, axis=0))
    recall_3 = np.mean(np.max(F_2, axis=1))
    f_measure_3 = util.f_measure(precision_3, recall_3)
    return f_measure_3, precision_3, recall_3


def first_n_three_layer_P(reference_patterns, estimated_patterns, n=5):
    """First n three-layer precision.

    This metric is basically the same as the three-layer FPR but it is only
    applied to the first n estimated patterns, and it only returns the
    precision. In MIREX and typically, n = 5.

    Examples
    --------
    >>> ref_patterns = mir_eval.io.load_patterns("ref_pattern.txt")
    >>> est_patterns = mir_eval.io.load_patterns("est_pattern.txt")
    >>> P = mir_eval.pattern.first_n_three_layer_P(ref_patterns,
    ...                                            est_patterns, n=5)

    Parameters
    ----------
    reference_patterns : list
        The reference patterns in the format returned by
        :func:`mir_eval.io.load_patterns()`
    estimated_patterns : list
        The estimated patterns in the same format
    n : int
        Number of patterns to consider from the estimated results, in
        the order they appear in the matrix
        (Default value = 5)

    Returns
    -------
    precision : float
        The first n three-layer Precision

    """

    validate(reference_patterns, estimated_patterns)
    # If no patterns were provided, metric is zero
    if _n_onset_midi(reference_patterns) == 0 or \
       _n_onset_midi(estimated_patterns) == 0:
        return 0., 0., 0.

    # Get only the first n patterns from the estimated results
    fn_est_patterns = estimated_patterns[:min(len(estimated_patterns), n)]

    # Compute the three-layer scores for the first n estimated patterns
    F, P, R = three_layer_FPR(reference_patterns, fn_est_patterns)

    return P    # Return the precision only


def first_n_target_proportion_R(reference_patterns, estimated_patterns, n=5):
    """First n target proportion establishment recall metric.

    This metric is similar is similar to the establishment FPR score, but it
    only takes into account the first n estimated patterns and it only
    outputs the Recall value of it.

    Examples
    --------
    >>> ref_patterns = mir_eval.io.load_patterns("ref_pattern.txt")
    >>> est_patterns = mir_eval.io.load_patterns("est_pattern.txt")
    >>> R = mir_eval.pattern.first_n_target_proportion_R(
    ...                                 ref_patterns, est_patterns, n=5)

    Parameters
    ----------
    reference_patterns : list
        The reference patterns in the format returned by
        :func:`mir_eval.io.load_patterns()`
    estimated_patterns : list
        The estimated patterns in the same format
    n : int
        Number of patterns to consider from the estimated results, in
        the order they appear in the matrix.
        (Default value = 5)

    Returns
    -------
    recall : float
        The first n target proportion Recall.

    """

    validate(reference_patterns, estimated_patterns)
    # If no patterns were provided, metric is zero
    if _n_onset_midi(reference_patterns) == 0 or \
       _n_onset_midi(estimated_patterns) == 0:
        return 0., 0., 0.

    # Get only the first n patterns from the estimated results
    fn_est_patterns = estimated_patterns[:min(len(estimated_patterns), n)]

    F, P, R = establishment_FPR(reference_patterns, fn_est_patterns)
    return R


def evaluate(ref_patterns, est_patterns, **kwargs):
    """Load data and perform the evaluation.

    Examples
    --------
    >>> ref_patterns = mir_eval.io.load_patterns("ref_pattern.txt")
    >>> est_patterns = mir_eval.io.load_patterns("est_pattern.txt")
    >>> scores = mir_eval.pattern.evaluate(ref_patterns, est_patterns)

    Parameters
    ----------
    ref_patterns : list
        The reference patterns in the format returned by
        :func:`mir_eval.io.load_patterns()`
    est_patterns : list
        The estimated patterns in the same format
    kwargs
        Additional keyword arguments which will be passed to the
        appropriate metric or preprocessing functions.

    Returns
    -------
    scores : dict
        Dictionary of scores, where the key is the metric name (str) and
        the value is the (float) score achieved.

    """

    # Compute all the metrics
    scores = collections.OrderedDict()

    # Standard scores
    scores['F'], scores['P'], scores['R'] = \
        util.filter_kwargs(standard_FPR, ref_patterns, est_patterns, **kwargs)

    # Establishment scores
    scores['F_est'], scores['P_est'], scores['R_est'] = \
        util.filter_kwargs(establishment_FPR, ref_patterns, est_patterns,
                           **kwargs)

    # Occurrence scores
    # Force these values for thresh
    kwargs['thresh'] = .5
    scores['F_occ.5'], scores['P_occ.5'], scores['R_occ.5'] = \
        util.filter_kwargs(occurrence_FPR, ref_patterns, est_patterns,
                           **kwargs)
    kwargs['thresh'] = .75
    scores['F_occ.75'], scores['P_occ.75'], scores['R_occ.75'] = \
        util.filter_kwargs(occurrence_FPR, ref_patterns, est_patterns,
                           **kwargs)

    # Three-layer scores
    scores['F_3'], scores['P_3'], scores['R_3'] = \
        util.filter_kwargs(three_layer_FPR, ref_patterns, est_patterns,
                           **kwargs)

    # First Five Patterns scores
    # Set default value of n
    if 'n' not in kwargs:
        kwargs['n'] = 5
    scores['FFP'] = util.filter_kwargs(first_n_three_layer_P, ref_patterns,
                                       est_patterns, **kwargs)
    scores['FFTP_est'] = \
        util.filter_kwargs(first_n_target_proportion_R, ref_patterns,
                           est_patterns, **kwargs)

    return scores



"""
Unit tests for mir_eval.chord
"""

import mir_eval
import numpy as np
import nose.tools
import warnings
import glob
import json

A_TOL = 1e-12

# Path to the fixture files
REF_GLOB = 'data/chord/ref*.lab'
EST_GLOB = 'data/chord/est*.lab'
SCORES_GLOB = 'data/chord/output*.json'


def __check_valid(function, parameters, result):
    ''' Helper function for checking the output of a function '''
    assert function(*parameters) == result


def __check_exception(function, parameters, exception):
    ''' Makes sure the provided function throws the provided
    exception given the provided input '''
    nose.tools.assert_raises(exception, function, *parameters)


def test_pitch_class_to_semitone():
    valid_classes = ['Gbb', 'G', 'G#', 'Cb', 'B#']
    valid_semitones = [5, 7, 8, 11, 0]

    for pitch_class, semitone in zip(valid_classes, valid_semitones):
        yield (__check_valid, mir_eval.chord.pitch_class_to_semitone,
               (pitch_class,), semitone)

    invalid_classes = ['Cab', '#C', 'bG']

    for pitch_class in invalid_classes:
        yield (__check_exception, mir_eval.chord.pitch_class_to_semitone,
               (pitch_class,), mir_eval.chord.InvalidChordException)


def test_scale_degree_to_semitone():
    valid_degrees = ['b7', '#3', '1', 'b1', '#7', 'bb5', '11', '#13']
    valid_semitones = [10, 5, 0, -1, 12, 5, 17, 22]

    for scale_degree, semitone in zip(valid_degrees, valid_semitones):
        yield (__check_valid, mir_eval.chord.scale_degree_to_semitone,
               (scale_degree,), semitone)

    invalid_degrees = ['7b', '4#', '77', '15']

    for scale_degree in invalid_degrees:
        yield (__check_exception, mir_eval.chord.scale_degree_to_semitone,
               (scale_degree,), mir_eval.chord.InvalidChordException)


def test_scale_degree_to_bitmap():

    def __check_bitmaps(function, parameters, result):
        actual = function(*parameters)
        assert np.all(actual == result), (actual, result)

    valid_degrees = ['3', '*3', 'b1', '9']
    valid_bitmaps = [[0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0],
                     [0, 0, 0, 0, -1, 0, 0, 0, 0, 0, 0, 0],
                     [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
                     [0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0]]

    for scale_degree, bitmap in zip(valid_degrees, valid_bitmaps):
        yield (__check_bitmaps, mir_eval.chord.scale_degree_to_bitmap,
               (scale_degree, True, 12), np.array(bitmap))

    yield (__check_bitmaps, mir_eval.chord.scale_degree_to_bitmap,
           ('9', False, 12), np.array([0] * 12))

    yield (__check_bitmaps, mir_eval.chord.scale_degree_to_bitmap,
           ('9', False, 15),
           np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1]))


def test_validate_chord_label():
    valid_labels = ['C', 'Eb:min/5', 'A#:dim7', 'B:maj(*1,*5)/3',
                    'A#:sus4', 'A:(9,11)']
    # For valid labels, calling the function without an error = pass
    for chord_label in valid_labels:
        yield (mir_eval.chord.validate_chord_label, chord_label)

    invalid_labels = ["C::maj", "C//5", "C((4)", "C5))",
                      "C:maj(*3/3", "Cmaj*3/3)", 'asdf']

    for chord_label in invalid_labels:
        yield (__check_exception, mir_eval.chord.validate_chord_label,
               (chord_label,), mir_eval.chord.InvalidChordException)


def test_split():
    labels = ['C', 'B:maj(*1,*3)/5', 'Ab:min/b3', 'N', 'G:(3)']
    splits = [['C', 'maj', set(), '1'],
              ['B', 'maj', set(['*1', '*3']), '5'],
              ['Ab', 'min', set(), 'b3'],
              ['N', '', set(), ''],
              ['G', '', set(['3']), '1']]

    for chord_label, split_chord in zip(labels, splits):
        yield (__check_valid, mir_eval.chord.split,
               (chord_label,), split_chord)

    # Test with reducing extended chords
    labels = ['C', 'C:minmaj7']
    splits = [['C', 'maj', set(), '1'],
              ['C', 'min', set(['7']), '1']]
    for chord_label, split_chord in zip(labels, splits):
        yield (__check_valid, mir_eval.chord.split,
               (chord_label, True), split_chord)

    # Test that an exception is raised when a chord with an omission but no
    # quality is supplied
    yield (__check_exception, mir_eval.chord.split,
           ('C(*5)',), mir_eval.chord.InvalidChordException)


def test_join():
    # Arguments are root, quality, extensions, bass
    splits = [('F#', '', None, ''),
              ('F#', 'hdim7', None, ''),
              ('F#', '', ['*b3', '4'], ''),
              ('F#', '', None, 'b7'),
              ('F#', '', ['*b3', '4'], 'b7'),
              ('F#', 'hdim7', None, 'b7'),
              ('F#', 'hdim7', ['*b3', '4'], 'b7')]
    labels = ['F#', 'F#:hdim7', 'F#:(*b3,4)', 'F#/b7',
              'F#:(*b3,4)/b7', 'F#:hdim7/b7', 'F#:hdim7(*b3,4)/b7']

    for split_chord, chord_label in zip(splits, labels):
        yield (__check_valid, mir_eval.chord.join,
               split_chord, chord_label)


def test_rotate_bitmaps_to_roots():
    def __check_bitmaps(bitmaps, roots, expected_bitmaps):
        ''' Helper function for checking bitmaps_to_roots '''
        ans = mir_eval.chord.rotate_bitmaps_to_roots(bitmaps, roots)
        assert np.all(ans == expected_bitmaps)

    bitmaps = [
        [1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0],
        [1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0],
        [1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0]]
    roots = [0, 5, 11]
    expected_bitmaps = [
        [1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0],
        [1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0],
        [0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1]]

    # The function can operate on many bitmaps/roots at a time
    # but we should only test them one at a time.
    for bitmap, root, expected_bitmap in zip(bitmaps, roots, expected_bitmaps):
        yield (__check_bitmaps, [bitmap], [root], [expected_bitmap])


def test_encode():
    def __check_encode(label, expected_root, expected_intervals,
                       expected_bass, reduce_extended_chords,
                       strict_bass_intervals):
        ''' Helper function for checking encode '''
        root, intervals, bass = mir_eval.chord.encode(
            label, reduce_extended_chords=reduce_extended_chords,
            strict_bass_intervals=strict_bass_intervals)
        assert root == expected_root, (root, expected_root)
        assert np.all(intervals == expected_intervals), (intervals,
                                                         expected_intervals)
        assert bass == expected_bass, (bass, expected_bass)

    labels = ['B:maj(*1,*3)/5', 'G:dim', 'C:(3)/3', 'A:9/b3']
    expected_roots = [11, 7, 0, 9]
    expected_intervals = [[0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
                          [1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0],
                          [1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0],
                          # Note that extended scale degrees are dropped.
                          [1, 0, 0, 1, 1, 0, 0, 1, 0, 0, 1, 0]]
    expected_bass = [7, 0, 4, 3]

    args = zip(labels, expected_roots, expected_intervals, expected_bass)
    for label, e_root, e_interval, e_bass in args:
        yield (__check_encode, label, e_root, e_interval, e_bass, False, False)

    # Non-chord bass notes *must* be explicitly named as extensions when
    #   strict_bass_intervals == True
    yield (__check_exception, mir_eval.chord.encode,
           ('G:dim(4)/6', False, True), mir_eval.chord.InvalidChordException)

    # Otherwise, we can cut a little slack.
    yield (__check_encode, 'G:dim(4)/6', 7,
                           [1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 0, 0], 9,
                           False, False)

    # Check that extended scale degrees are mapped back into pitch classes.
    yield (__check_encode, 'A:9', 9,
                           [1, 0, 1, 0, 1, 0, 0, 1, 0, 0, 1, 0], 0,
                           True, False)


def test_encode_many():
    def __check_encode_many(labels, expected_roots, expected_intervals,
                            expected_basses):
        ''' Does all of the logic for checking encode_many '''
        roots, intervals, basses = mir_eval.chord.encode_many(labels)
        assert np.all(roots == expected_roots)
        assert np.all(intervals == expected_intervals)
        assert np.all(basses == expected_basses)

    labels = ['B:maj(*1,*3)/5',
              'B:maj(*1,*3)/5',
              'N',
              'C:min',
              'C:min']
    expected_roots = [11, 11, -1, 0, 0]
    expected_intervals = [
        [0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        [1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0],
        [1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0]]
    expected_basses = [7, 7, -1, 0, 0]

    yield (__check_encode_many, labels, expected_roots, expected_intervals,
           expected_basses)


def __check_one_metric(metric, ref_label, est_label, score):
    ''' Checks that a metric function produces score given ref_label and
    est_label '''
    # We provide a dummy interval.  We're just checking one pair
    # of labels at a time.
    assert metric([ref_label], [est_label]) == score


def __check_not_comparable(metric, ref_label, est_label):
    ''' Checks that ref_label is not comparable to est_label by metric '''
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter('always')
        # Try to produce the warning
        score = mir_eval.chord.weighted_accuracy(metric([ref_label],
                                                        [est_label]),
                                                 np.array([1]))
        assert len(w) == 1
        assert issubclass(w[-1].category, UserWarning)
        assert str(w[-1].message) == ("No reference chords were comparable "
                                      "to estimated chords, returning 0.")
        # And confirm that the metric is 0
        assert np.allclose(score, 0)

# TODO(ejhumphrey): Comparison functions lacking unit tests.
# test_root()


def test_mirex():
    ref_labels = ['N', 'C:maj', 'C:maj', 'C:maj', 'C:min', 'C:maj',
                  'C:maj',  'G:min',  'C:maj', 'C:min',   'C:min',
                  'C:maj',  'F:maj',  'C:maj7',    'A:maj', 'A:maj']
    est_labels = ['N', 'N',     'C:aug', 'C:dim', 'C:dim', 'C:5',
                  'C:sus4', 'G:sus2', 'G:maj', 'C:hdim7', 'C:min7',
                  'C:maj6', 'F:min6', 'C:minmaj7', 'A:7',   'A:9']
    scores = [1.0, 0.0, 0.0, 0.0, 0.0, 0.0,
              0.0, 0.0, 0.0, 0.0, 1.0,
              1.0, 0.0, 1.0, 1.0, 1.0]

    for ref_label, est_label, score in zip(ref_labels, est_labels, scores):
        yield (__check_one_metric, mir_eval.chord.mirex,
               ref_label, est_label, score)

    ref_not_comparable = ['C:5', 'X']
    est_not_comparable = ['C:maj', 'N']

    for ref_label, est_label in zip(ref_not_comparable, est_not_comparable):
        yield (__check_not_comparable, mir_eval.chord.mirex,
               ref_label, est_label)


def test_thirds():
    ref_labels = ['N', 'C:maj', 'C:maj', 'C:maj', 'C:min',
                  'C:maj', 'G:min', 'C:maj', 'C:min', 'C:min',
                  'C:maj', 'F:maj', 'C:maj', 'A:maj', 'A:maj']
    est_labels = ['N', 'N', 'C:aug', 'C:dim', 'C:dim',
                  'C:sus4', 'G:sus2', 'G:maj', 'C:hdim7', 'C:min7',
                  'C:maj6', 'F:min6', 'C:minmaj7', 'A:7', 'A:9']
    scores = [1.0, 0.0, 1.0, 0.0, 1.0,
              1.0, 0.0, 0.0, 1.0, 1.0,
              1.0, 0.0, 0.0, 1.0, 1.0]

    for ref_label, est_label, score in zip(ref_labels, est_labels, scores):
        yield (__check_one_metric, mir_eval.chord.thirds,
               ref_label, est_label, score)

    yield (__check_not_comparable, mir_eval.chord.thirds, 'X', 'N')


def test_thirds_inv():
    ref_labels = ['C:maj/5',  'G:min',    'C:maj',   'C:min/b3',   'C:min']
    est_labels = ['C:sus4/5', 'G:min/b3', 'C:maj/5', 'C:hdim7/b3', 'C:dim']
    scores = [1.0, 0.0, 0.0, 1.0, 1.0]

    for ref_label, est_label, score in zip(ref_labels, est_labels, scores):
        yield (__check_one_metric, mir_eval.chord.thirds_inv,
               ref_label, est_label, score)

    yield (__check_not_comparable, mir_eval.chord.thirds_inv, 'X', 'N')


def test_triads():
    ref_labels = ['C:min',  'C:maj', 'C:maj', 'C:min', 'C:maj',
                  'C:maj',  'G:min',     'C:maj', 'C:min',   'C:min']
    est_labels = ['C:min7', 'C:7',   'C:aug', 'C:dim', 'C:sus2',
                  'C:sus4', 'G:minmaj7', 'G:maj', 'C:hdim7', 'C:min6']
    scores = [1.0, 1.0, 0.0, 0.0, 0.0,
              0.0, 1.0, 0.0, 0.0, 1.0]

    for ref_label, est_label, score in zip(ref_labels, est_labels, scores):
        yield (__check_one_metric, mir_eval.chord.triads,
               ref_label, est_label, score)

    yield (__check_not_comparable, mir_eval.chord.triads, 'X', 'N')


def test_triads_inv():
    ref_labels = ['C:maj/5',  'G:min',    'C:maj', 'C:min/b3',  'C:min/b3']
    est_labels = ['C:maj7/5', 'G:min7/5', 'C:7/5', 'C:min6/b3', 'C:dim/b3']
    scores = [1.0, 0.0, 0.0, 1.0, 0.0]

    for ref_label, est_label, score in zip(ref_labels, est_labels, scores):
        yield (__check_one_metric, mir_eval.chord.triads_inv,
               ref_label, est_label, score)

    yield (__check_not_comparable, mir_eval.chord.triads_inv, 'X', 'N')


def test_tetrads():
    ref_labels = ['C:min', 'C:maj', 'C:7', 'C:maj7', 'C:sus2',
                  'C:7/3', 'G:min', 'C:maj', 'C:min', 'C:min']
    est_labels = ['C:min7', 'C:maj6', 'C:9', 'C:maj7/5', 'C:sus2/2',
                  'C:11/b7', 'G:sus2', 'G:maj', 'C:hdim7', 'C:minmaj7']
    scores = [0.0, 0.0, 1.0, 1.0, 1.0,
              1.0, 0.0, 0.0, 0.0, 0.0]

    for ref_label, est_label, score in zip(ref_labels, est_labels, scores):
        yield (__check_one_metric, mir_eval.chord.tetrads,
               ref_label, est_label, score)

    yield (__check_not_comparable, mir_eval.chord.tetrads, 'X', 'N')


def test_tetrads_inv():
    ref_labels = ['C:maj7/5', 'G:min', 'C:7/5', 'C:min/b3', 'C:min9']
    est_labels = ['C:maj7/3', 'G:min/b3', 'C:13/5', 'C:hdim7/b3', 'C:min7']
    scores = [0.0, 0.0, 1.0, 0.0, 1.0]

    for ref_label, est_label, score in zip(ref_labels, est_labels, scores):
        yield (__check_one_metric, mir_eval.chord.tetrads_inv,
               ref_label, est_label, score)

    yield (__check_not_comparable, mir_eval.chord.tetrads_inv, 'X', 'N')


def test_majmin():
    ref_labels = ['N', 'C:maj', 'C:maj', 'C:min', 'G:maj7']
    est_labels = ['N', 'N',     'C:aug', 'C:dim', 'G']
    scores = [1.0,  0.0, 0.0, 0.0, 1.0]

    for ref_label, est_label, score in zip(ref_labels, est_labels, scores):
        yield (__check_one_metric, mir_eval.chord.majmin,
               ref_label, est_label, score)

    ref_not_comparable = ['C:aug', 'X']
    est_not_comparable = ['C:maj', 'N']

    for ref_label, est_label in zip(ref_not_comparable, est_not_comparable):
        yield (__check_not_comparable, mir_eval.chord.majmin,
               ref_label, est_label)


def test_majmin_inv():
    ref_labels = ['C:maj/5',  'G:min',    'C:maj/5', 'C:min7',
                  'G:min/b3', 'C:maj7/5', 'C:7']
    est_labels = ['C:sus4/5', 'G:min/b3', 'C:maj/5', 'C:min',
                  'G:min/b3', 'C:maj/5', 'C:maj']
    scores = [0.0, 0.0, 1.0, 1.0,
              1.0, 1.0, 1.0]

    for ref_label, est_label, score in zip(ref_labels, est_labels, scores):
        yield (__check_one_metric, mir_eval.chord.majmin_inv,
               ref_label, est_label, score)

    ref_not_comparable = ['C:hdim7/b3', 'C:maj/4', 'C:maj/2', 'X']
    est_not_comparable = ['C:min/b3', 'C:maj/4', 'C:sus2/2',  'N']

    for ref_label, est_label in zip(ref_not_comparable, est_not_comparable):
        yield (__check_not_comparable, mir_eval.chord.majmin_inv,
               ref_label, est_label)


def test_sevenths():
    ref_labels = ['C:min',  'C:maj',  'C:7', 'C:maj7',
                  'C:7/3',   'G:min',  'C:maj', 'C:7']
    est_labels = ['C:min7', 'C:maj6', 'C:9', 'C:maj7/5',
                  'C:11/b7', 'G:sus2', 'G:maj', 'C:maj7']
    scores = [0.0, 0.0, 1.0, 1.0,
              1.0, 0.0, 0.0, 0.0]

    for ref_label, est_label, score in zip(ref_labels, est_labels, scores):
        yield (__check_one_metric, mir_eval.chord.sevenths,
               ref_label, est_label, score)

    ref_not_comparable = ['C:sus2',   'C:hdim7', 'X']
    est_not_comparable = ['C:sus2/2', 'C:hdim7', 'N']
    for ref_label, est_label in zip(ref_not_comparable, est_not_comparable):
        yield (__check_not_comparable, mir_eval.chord.sevenths,
               ref_label, est_label)


def test_sevenths_inv():
    ref_labels = ['C:maj7/5', 'G:min',    'C:7/5',  'C:min7/b7']
    est_labels = ['C:maj7/3', 'G:min/b3', 'C:13/5', 'C:min7/b7']
    scores = [0.0, 0.0, 1.0, 1.0]

    for ref_label, est_label, score in zip(ref_labels, est_labels, scores):
        yield (__check_one_metric, mir_eval.chord.sevenths_inv,
               ref_label, est_label, score)

    ref_not_comparable = ['C:dim7/b3', 'X']
    est_not_comparable = ['C:dim7/b3', 'N']
    for ref_label, est_label in zip(ref_not_comparable, est_not_comparable):
        yield (__check_not_comparable, mir_eval.chord.sevenths_inv,
               ref_label, est_label)


def test_directional_hamming_distance():
    ref_ivs = np.array([[0., 1.], [1., 2.], [2., 3.]])
    est_ivs = np.array([[0., 0.9], [0.9, 1.8], [1.8, 2.5]])
    dhd_ref_to_est = (0.1 + 0.2 + 0.5) / 3.
    dhd_est_to_ref = (0.0 + 0.1 + 0.2) / 2.5

    dhd = mir_eval.chord.directional_hamming_distance
    assert np.allclose(dhd_ref_to_est, dhd(ref_ivs, est_ivs))
    assert np.allclose(dhd_est_to_ref, dhd(est_ivs, ref_ivs))
    assert np.allclose(0, dhd(ref_ivs, ref_ivs))
    assert np.allclose(0, dhd(est_ivs, est_ivs))

    ivs_overlap_all = np.array([[0., 1.], [0.9, 2.]])
    ivs_overlap_one = np.array([[0., 1.], [0.9, 2.], [2., 3.]])
    nose.tools.assert_raises(ValueError, dhd, ivs_overlap_all, est_ivs)
    nose.tools.assert_raises(ValueError, dhd, ivs_overlap_one, est_ivs)


def test_segmentation_functions():
    ref_ivs = np.array([[0., 2.], [2., 2.5], [2.5, 3.2]])
    est_ivs = np.array([[0., 3.], [3., 3.5]])
    true_oseg = 1. - 0.2 / 3.2
    true_useg = 1. - (1. + 0.2) / 3.5
    true_seg = min(true_oseg, true_useg)
    assert np.allclose(true_oseg, mir_eval.chord.overseg(ref_ivs, est_ivs))
    assert np.allclose(true_useg, mir_eval.chord.underseg(ref_ivs, est_ivs))
    assert np.allclose(true_seg, mir_eval.chord.seg(ref_ivs, est_ivs))

    ref_ivs = np.array([[0., 2.], [2., 2.5], [2.5, 3.2]])
    est_ivs = np.array([[0., 2.], [2., 2.5], [2.5, 3.2]])
    true_oseg = 1.0
    true_useg = 1.0
    true_seg = 1.0
    assert np.allclose(true_oseg, mir_eval.chord.overseg(ref_ivs, est_ivs))
    assert np.allclose(true_useg, mir_eval.chord.underseg(ref_ivs, est_ivs))
    assert np.allclose(true_seg, mir_eval.chord.seg(ref_ivs, est_ivs))

    ref_ivs = np.array([[0., 2.], [2., 2.5], [2.5, 3.2]])
    est_ivs = np.array([[0., 3.2]])
    true_oseg = 1.0
    true_useg = 1 - 1.2 / 3.2
    true_seg = min(true_oseg, true_useg)
    assert np.allclose(true_oseg, mir_eval.chord.overseg(ref_ivs, est_ivs))
    assert np.allclose(true_useg, mir_eval.chord.underseg(ref_ivs, est_ivs))
    assert np.allclose(true_seg, mir_eval.chord.seg(ref_ivs, est_ivs))

    ref_ivs = np.array([[0., 2.], [2., 2.5], [2.5, 3.2]])
    est_ivs = np.array([[3.2, 3.5]])
    true_oseg = 1.0
    true_useg = 1.0
    true_seg = 1.0
    assert np.allclose(true_oseg, mir_eval.chord.overseg(ref_ivs, est_ivs))
    assert np.allclose(true_useg, mir_eval.chord.underseg(ref_ivs, est_ivs))
    assert np.allclose(true_seg, mir_eval.chord.seg(ref_ivs, est_ivs))


def test_merge_chord_intervals():
    intervals = np.array([[0., 1.], [1., 2.], [2., 3], [3., 4.], [4., 5.]])
    labels = ['C:maj', 'C:(1,3,5)', 'A:maj', 'A:maj7', 'A:maj7/3']
    assert np.allclose(np.array([[0., 2.], [2., 3], [3., 4.], [4., 5.]]),
                       mir_eval.chord.merge_chord_intervals(intervals, labels))


def test_weighted_accuracy():
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter('always')
        # First, test for a warning on empty beats
        score = mir_eval.chord.weighted_accuracy(np.array([1, 0, 1]),
                                                 np.array([0, 0, 0]))
        assert len(w) == 1
        assert issubclass(w[-1].category, UserWarning)
        assert str(w[-1].message) == 'No nonzero weights, returning 0'
        # And that the metric is 0
        assert np.allclose(score, 0)

    # len(comparisons) must equal len(weights)
    comparisons = np.array([1, 0, 1])
    weights = np.array([1, 1])
    nose.tools.assert_raises(ValueError, mir_eval.chord.weighted_accuracy,
                             comparisons, weights)
    # Weights must all be positive
    comparisons = np.array([1, 1])
    weights = np.array([-1, -1])
    nose.tools.assert_raises(ValueError, mir_eval.chord.weighted_accuracy,
                             comparisons, weights)

    # Make sure accuracy = 1 and 0 when all comparisons are True and False resp
    comparisons = np.array([1, 1, 1])
    weights = np.array([1, 1, 1])
    score = mir_eval.chord.weighted_accuracy(comparisons, weights)
    assert np.allclose(score, 1)
    comparisons = np.array([0, 0, 0])
    score = mir_eval.chord.weighted_accuracy(comparisons, weights)
    assert np.allclose(score, 0)


def __check_score(sco_f, metric, score, expected_score):
    assert np.allclose(score, expected_score, atol=A_TOL)


def test_chord_functions():
    # Load in all files in the same order
    ref_files = sorted(glob.glob(REF_GLOB))
    est_files = sorted(glob.glob(EST_GLOB))
    sco_files = sorted(glob.glob(SCORES_GLOB))

    assert len(ref_files) == len(est_files) == len(sco_files) > 0

    # Regression tests
    for ref_f, est_f, sco_f in zip(ref_files, est_files, sco_files):
        with open(sco_f, 'r') as f:
            expected_scores = json.load(f)
        # Load in an example beat annotation
        ref_intervals, ref_labels = mir_eval.io.load_labeled_intervals(ref_f)
        # Load in an example beat tracker output
        est_intervals, est_labels = mir_eval.io.load_labeled_intervals(est_f)
        # Compute scores
        scores = mir_eval.chord.evaluate(ref_intervals, ref_labels,
                                         est_intervals, est_labels)
        # Compare them
        for metric in scores:
            # This is a simple hack to make nosetest's messages more useful
            yield (__check_score, sco_f, metric, scores[metric],
                   expected_scores[metric])


def test_quality_to_bitmap():

    # Test simple case
    assert np.all(mir_eval.chord.quality_to_bitmap('maj') == np.array(
        [1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0]))

    # Check exceptions for qualities not in the QUALITIES list
    invalid_qualities = ['maj5', '2', '#7']
    for quality in invalid_qualities:
        yield (__check_exception, mir_eval.chord.quality_to_bitmap,
               (quality,), mir_eval.chord.InvalidChordException)


def test_validate():
    # Test that the validate function raises the appropriate errors and
    # warnings
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter('always')
        # First, test for warnings on empty labels
        mir_eval.chord.validate([], [])
        assert len(w) == 2
        assert issubclass(w[-1].category, UserWarning)
        assert str(w[-1].message) == "Estimated labels are empty"
        assert issubclass(w[-2].category, UserWarning)
        assert str(w[-2].message) == "Reference labels are empty"
        # Test that error is thrown on different-length labels
        nose.tools.assert_raises(
            ValueError, mir_eval.chord.validate, [], ['C'])



'''
The goal of a tempo estimation algorithm is to automatically detect the tempo
of a piece of music, measured in beats per minute (BPM).

See http://www.music-ir.org/mirex/wiki/2014:Audio_Tempo_Estimation for a
description of the task and evaluation criteria.

Conventions
-----------

Reference and estimated tempi should be positive, and provided in ascending
order as a numpy array of length 2.

The weighting value from the reference must be a float in the range [0, 1].

Metrics
-------
* :func:`mir_eval.tempo.detection`: Relative error, hits, and weighted
  precision of tempo estimation.

'''

import warnings
import numpy as np
import collections
from . import util


def validate_tempi(tempi, reference=True):
    """Checks that there are two non-negative tempi.
    For a reference value, at least one tempo has to be greater than zero.

    Parameters
    ----------
    tempi : np.ndarray
        length-2 array of tempo, in bpm

    reference : bool
        indicates a reference value

    """

    if tempi.size != 2:
        raise ValueError('tempi must have exactly two values')

    if not np.all(np.isfinite(tempi)) or np.any(tempi < 0):
        raise ValueError('tempi={} must be non-negative numbers'.format(tempi))

    if reference and np.all(tempi == 0):
        raise ValueError('reference tempi={} must have one'
                         ' value greater than zero'.format(tempi))


def validate(reference_tempi, reference_weight, estimated_tempi):
    """Checks that the input annotations to a metric look like valid tempo
    annotations.

    Parameters
    ----------
    reference_tempi : np.ndarray
        reference tempo values, in bpm

    reference_weight : float
        perceptual weight of slow vs fast in reference

    estimated_tempi : np.ndarray
        estimated tempo values, in bpm

    """
    validate_tempi(reference_tempi, reference=True)
    validate_tempi(estimated_tempi, reference=False)

    if reference_weight < 0 or reference_weight > 1:
        raise ValueError('Reference weight must lie in range [0, 1]')


def detection(reference_tempi, reference_weight, estimated_tempi, tol=0.08):
    """Compute the tempo detection accuracy metric.

    Parameters
    ----------
    reference_tempi : np.ndarray, shape=(2,)
        Two non-negative reference tempi

    reference_weight : float > 0
        The relative strength of ``reference_tempi[0]`` vs
        ``reference_tempi[1]``.

    estimated_tempi : np.ndarray, shape=(2,)
        Two non-negative estimated tempi.

    tol : float in [0, 1]:
        The maximum allowable deviation from a reference tempo to
        count as a hit.
        ``|est_t - ref_t| <= tol * ref_t``
        (Default value = 0.08)

    Returns
    -------
    p_score : float in [0, 1]
        Weighted average of recalls:
        ``reference_weight * hits[0] + (1 - reference_weight) * hits[1]``

    one_correct : bool
        True if at least one reference tempo was correctly estimated

    both_correct : bool
        True if both reference tempi were correctly estimated

    Raises
    ------
    ValueError
        If the input tempi are ill-formed

        If the reference weight is not in the range [0, 1]

        If ``tol < 0`` or ``tol > 1``.
    """

    validate(reference_tempi, reference_weight, estimated_tempi)

    if tol < 0 or tol > 1:
        raise ValueError('invalid tolerance {}: must lie in the range '
                         '[0, 1]'.format(tol))
    if tol == 0.:
        warnings.warn('A tolerance of 0.0 may not '
                      'lead to the results you expect.')

    hits = [False, False]

    for i, ref_t in enumerate(reference_tempi):
        if ref_t > 0:
            # Compute the relative error for this reference tempo
            f_ref_t = float(ref_t)
            relative_error = np.min(np.abs(ref_t - estimated_tempi) / f_ref_t)

            # Count the hits
            hits[i] = relative_error <= tol

    p_score = reference_weight * hits[0] + (1.0-reference_weight) * hits[1]

    one_correct = bool(np.max(hits))
    both_correct = bool(np.min(hits))

    return p_score, one_correct, both_correct


def evaluate(reference_tempi, reference_weight, estimated_tempi, **kwargs):
    """Compute all metrics for the given reference and estimated annotations.

    Parameters
    ----------
    reference_tempi : np.ndarray, shape=(2,)
        Two non-negative reference tempi

    reference_weight : float > 0
        The relative strength of ``reference_tempi[0]`` vs
        ``reference_tempi[1]``.

    estimated_tempi : np.ndarray, shape=(2,)
        Two non-negative estimated tempi.

    kwargs
        Additional keyword arguments which will be passed to the
        appropriate metric or preprocessing functions.

    Returns
    -------
    scores : dict
        Dictionary of scores, where the key is the metric name (str) and
        the value is the (float) score achieved.
    """
    # Compute all metrics
    scores = collections.OrderedDict()

    (scores['P-score'],
     scores['One-correct'],
     scores['Both-correct']) = util.filter_kwargs(detection, reference_tempi,
                                                  reference_weight,
                                                  estimated_tempi,
                                                  **kwargs)

    return scores



'''
The goal of an onset detection algorithm is to automatically determine when
notes are played in a piece of music.  The primary method used to evaluate
onset detectors is to first determine which estimated onsets are "correct",
where correctness is defined as being within a small window of a reference
onset.

Based in part on this script:

    https://github.com/CPJKU/onset_detection/blob/master/onset_evaluation.py

Conventions
-----------

Onsets should be provided in the form of a 1-dimensional array of onset
times in seconds in increasing order.

Metrics
-------

* :func:`mir_eval.onset.f_measure`: Precision, Recall, and F-measure scores
  based on the number of esimated onsets which are sufficiently close to
  reference onsets.
'''

import collections
from . import util
import warnings


# The maximum allowable beat time
MAX_TIME = 30000.


def validate(reference_onsets, estimated_onsets):
    """Checks that the input annotations to a metric look like valid onset time
    arrays, and throws helpful errors if not.

    Parameters
    ----------
    reference_onsets : np.ndarray
        reference onset locations, in seconds
    estimated_onsets : np.ndarray
        estimated onset locations, in seconds

    """
    # If reference or estimated onsets are empty, warn because metric will be 0
    if reference_onsets.size == 0:
        warnings.warn("Reference onsets are empty.")
    if estimated_onsets.size == 0:
        warnings.warn("Estimated onsets are empty.")
    for onsets in [reference_onsets, estimated_onsets]:
        util.validate_events(onsets, MAX_TIME)


def f_measure(reference_onsets, estimated_onsets, window=.05):
    """Compute the F-measure of correct vs incorrectly predicted onsets.
    "Corectness" is determined over a small window.

    Examples
    --------
    >>> reference_onsets = mir_eval.io.load_events('reference.txt')
    >>> estimated_onsets = mir_eval.io.load_events('estimated.txt')
    >>> F, P, R = mir_eval.onset.f_measure(reference_onsets,
    ...                                    estimated_onsets)

    Parameters
    ----------
    reference_onsets : np.ndarray
        reference onset locations, in seconds
    estimated_onsets : np.ndarray
        estimated onset locations, in seconds
    window : float
        Window size, in seconds
        (Default value = .05)

    Returns
    -------
    f_measure : float
        2*precision*recall/(precision + recall)
    precision : float
        (# true positives)/(# true positives + # false positives)
    recall : float
        (# true positives)/(# true positives + # false negatives)

    """
    validate(reference_onsets, estimated_onsets)
    # If either list is empty, return 0s
    if reference_onsets.size == 0 or estimated_onsets.size == 0:
        return 0., 0., 0.
    # Compute the best-case matching between reference and estimated onset
    # locations
    matching = util.match_events(reference_onsets, estimated_onsets, window)

    precision = float(len(matching))/len(estimated_onsets)
    recall = float(len(matching))/len(reference_onsets)
    # Compute F-measure and return all statistics
    return util.f_measure(precision, recall), precision, recall


def evaluate(reference_onsets, estimated_onsets, **kwargs):
    """Compute all metrics for the given reference and estimated annotations.

    Examples
    --------
    >>> reference_onsets = mir_eval.io.load_events('reference.txt')
    >>> estimated_onsets = mir_eval.io.load_events('estimated.txt')
    >>> scores = mir_eval.onset.evaluate(reference_onsets,
    ...                                  estimated_onsets)

    Parameters
    ----------
    reference_onsets : np.ndarray
        reference onset locations, in seconds
    estimated_onsets : np.ndarray
        estimated onset locations, in seconds
    kwargs
        Additional keyword arguments which will be passed to the
        appropriate metric or preprocessing functions.

    Returns
    -------
    scores : dict
        Dictionary of scores, where the key is the metric name (str) and
        the value is the (float) score achieved.

    """
    # Compute all metrics
    scores = collections.OrderedDict()

    (scores['F-measure'],
     scores['Precision'],
     scores['Recall']) = util.filter_kwargs(f_measure, reference_onsets,
                                            estimated_onsets, **kwargs)

    return scores



'''
unit tests for mir_eval.separation

load randomly generated source and estimated source signals and
the output from BSS_eval MATLAB implementation, make sure the results
from mir_eval numerically match.
'''

import numpy as np
import mir_eval
import glob
import nose.tools
import json
import os
import warnings

A_TOL = 1e-2

REF_GLOB = 'data/separation/ref*'
EST_GLOB = 'data/separation/est*'
SCORES_GLOB = 'data/separation/output*.json'


def __load_and_stack_wavs(directory):
    ''' Load all wavs in a directory and stack them vertically into a matrix
    '''
    stacked_audio_data = []
    global_fs = None
    for f in sorted(glob.glob(os.path.join(directory, '*.wav'))):
        audio_data, fs = mir_eval.io.load_wav(f)
        assert global_fs is None or fs == global_fs
        global_fs = fs
        stacked_audio_data.append(audio_data)
    return np.vstack(stacked_audio_data)


def __generate_multichannel(mono_sig, nchan=2, gain=1.0, reverse=False):
    ''' Turn a single channel (ie. mono) audio sample into a multichannel
    (e.g. stereo)
    Note: to achieve channels of silence pass gain=0
    '''
    # add the channels dimension
    input_3d = np.atleast_3d(mono_sig)
    # get the desired number of channels
    stackin = [input_3d]*nchan
    # apply the gain to the new channels
    stackin[1:] = np.multiply(gain, stackin[1:])
    if reverse:
        # reverse the new channels
        stackin[1:] = stackin[1:][:][::-1]
    return np.dstack(stackin)


def __check_score(sco_f, metric, score, expected_score):
    assert np.allclose(score, expected_score, atol=A_TOL)


def __unit_test_empty_input(metric):
    if (metric == mir_eval.separation.bss_eval_sources or
            metric == mir_eval.separation.bss_eval_images):
        args = [np.array([]), np.array([])]
    elif (metric == mir_eval.separation.bss_eval_sources_framewise or
            metric == mir_eval.separation.bss_eval_images_framewise):
        args = [np.array([]), np.array([]), 40, 20]
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter('always')
        # First, test for a warning on empty audio data
        metric(*args)
        assert len(w) == 2
        assert issubclass(w[-1].category, UserWarning)
        assert str(w[-1].message) == ("estimated_sources is empty, "
                                      "should be of size (nsrc, nsample).  "
                                      "sdr, sir, sar, and perm will all be "
                                      "empty np.ndarrays")
        # And that the metric returns empty arrays
        assert np.allclose(metric(*args), np.array([]))


def __unit_test_silent_input(metric):
    # Test for error when there is a silent reference/estimated source
    if (metric == mir_eval.separation.bss_eval_images or
            metric == mir_eval.separation.bss_eval_images_framewise):
        ref_sources = np.vstack((np.zeros((1, 100, 2)),
                                 np.random.random_sample((2, 100, 2))))
        est_sources = np.vstack((np.zeros((1, 100, 2)),
                                 np.random.random_sample((2, 100, 2))))
    else:
        ref_sources = np.vstack((np.zeros(100),
                                 np.random.random_sample((2, 100))))
        est_sources = np.vstack((np.zeros(100),
                                 np.random.random_sample((2, 100))))
    if (metric == mir_eval.separation.bss_eval_sources or
            metric == mir_eval.separation.bss_eval_images):
        nose.tools.assert_raises(ValueError, metric, ref_sources[:2],
                                 est_sources[1:])
        nose.tools.assert_raises(ValueError, metric, ref_sources[1:],
                                 est_sources[:2])
    elif (metric == mir_eval.separation.bss_eval_sources_framewise or
            metric == mir_eval.separation.bss_eval_images_framewise):
        nose.tools.assert_raises(ValueError, metric, ref_sources[:2],
                                 est_sources[1:], 40, 20)
        nose.tools.assert_raises(ValueError, metric, ref_sources[1:],
                                 est_sources[:2], 40, 20)
    else:
        raise ValueError('Unknown metric {}'.format(metric))


def __unit_test_partial_silence(metric):
    # Test for a full window of silence in reference/estimated source
    if metric == mir_eval.separation.bss_eval_sources_framewise:
        silence = np.zeros((2, 20))
        sound = np.random.random_sample((2, 20))
    elif metric == mir_eval.separation.bss_eval_images_framewise:
        silence = np.zeros((2, 20, 2))
        sound = np.random.random_sample((2, 20, 2))
    else:
        raise ValueError('Unknown metric {}'.format(metric))
    # test with silence in the reference
    results = metric(np.concatenate((sound, silence, sound),
                                    axis=1),
                     np.concatenate((sound, sound, sound),
                                    axis=1),
                     window=10,
                     hop=10)
    for measure in results:
        for idx, source in enumerate(measure):
            if idx < 2 or idx > 3:
                assert not np.isnan(source[idx])
            elif idx < 4:
                assert np.isnan(source[idx])
            else:
                raise ValueError('Testing error in partial silence test')
    # test with silence in the estimate
    results = metric(np.concatenate((sound, sound, sound),
                                    axis=1),
                     np.concatenate((sound, silence, sound),
                                    axis=1),
                     window=10,
                     hop=10)
    for measure in results:
        for idx, source in enumerate(measure):
            if idx < 2 or idx > 3:
                assert not np.isnan(source[idx])
            elif idx < 4:
                assert np.isnan(source[idx])
            else:
                raise ValueError('Testing error in partial silence test')


def __unit_test_incompatible_shapes(metric):
    # Test for error when shape is different
    if (metric == mir_eval.separation.bss_eval_images or
            metric == mir_eval.separation.bss_eval_images_framewise):
        sources_4 = np.random.random_sample((4, 100, 2))
        sources_3 = np.random.random_sample((3, 100, 2))
        sources_4_chan = np.random.random_sample((4, 100, 3))
    else:
        sources_4 = np.random.random_sample((4, 100))
        sources_3 = np.random.random_sample((3, 100))
    if (metric == mir_eval.separation.bss_eval_sources or
            metric == mir_eval.separation.bss_eval_images):
        args1 = [sources_3, sources_4]
        args2 = [sources_4, sources_3]
    elif (metric == mir_eval.separation.bss_eval_sources_framewise or
            metric == mir_eval.separation.bss_eval_images_framewise):
        args1 = [sources_3, sources_4, 40, 20]
        args2 = [sources_4, sources_3, 40, 20]
    else:
        raise ValueError('Unknown metric {}'.format(metric))
    nose.tools.assert_raises(ValueError, metric, *args1)
    nose.tools.assert_raises(ValueError, metric, *args2)
    if (metric == mir_eval.separation.bss_eval_images or
            metric == mir_eval.separation.bss_eval_images_framewise):
        nose.tools.assert_raises(ValueError, metric, sources_4, sources_4_chan)


def __unit_test_too_many_sources(metric):
    # Test for error when too many sources or references are provided
    many_sources = np.random.random_sample((mir_eval.separation.MAX_SOURCES*2,
                                            400))
    if metric == mir_eval.separation.bss_eval_sources:
        nose.tools.assert_raises(ValueError, metric, many_sources,
                                 many_sources)
    elif metric == mir_eval.separation.bss_eval_sources_framewise:
        nose.tools.assert_raises(ValueError, metric, many_sources,
                                 many_sources, 40, 20)


def __unit_test_too_many_dimensions(metric):
    # Test for detection of too high dimensioned images
    ref_sources = np.random.random_sample((4, 100, 2, 3))
    est_sources = np.random.random_sample((4, 100, 2, 3))
    nose.tools.assert_raises(ValueError, metric, ref_sources, est_sources)


def __unit_test_default_permutation(metric):
    # Test for default permutation matrix when not computing permutation
    if metric == mir_eval.separation.bss_eval_sources:
        ref_sources = np.random.random_sample((4, 100))
        est_sources = np.random.random_sample((4, 100))
    elif metric == mir_eval.separation.bss_eval_images:
        ref_sources = np.random.random_sample((4, 100, 2))
        est_sources = np.random.random_sample((4, 100, 2))
    else:
        raise ValueError('Unknown metric {}'.format(metric))
    results = metric(ref_sources, est_sources, compute_permutation=False)
    assert np.array_equal(results[-1], np.asarray([0, 1, 2, 3]))


def __unit_test_framewise_small_window(metric):
    # Test for invalid win/hop parameter detection
    if metric == mir_eval.separation.bss_eval_sources_framewise:
        ref_sources = np.random.random_sample((4, 100))
        est_sources = np.random.random_sample((4, 100))
        comparison_fcn = mir_eval.separation.bss_eval_sources
    elif metric == mir_eval.separation.bss_eval_images_framewise:
        ref_sources = np.random.random_sample((4, 100, 2))
        est_sources = np.random.random_sample((4, 100, 2))
        comparison_fcn = mir_eval.separation.bss_eval_images
    else:
        raise ValueError('Unknown metric {}'.format(metric))
    # Test with window larger than source length
    assert np.allclose(np.squeeze(metric(ref_sources,
                                         est_sources,
                                         window=120,
                                         hop=20)),
                       comparison_fcn(ref_sources, est_sources, False),
                       atol=A_TOL)
    # Test with hop larger than source length
    assert np.allclose(np.squeeze(metric(ref_sources,
                                         est_sources,
                                         window=20,
                                         hop=120)),
                       comparison_fcn(ref_sources, est_sources, False),
                       atol=A_TOL)


def test_separation_functions():
    # Load in all files in the same order
    ref_files = sorted(glob.glob(REF_GLOB))
    est_files = sorted(glob.glob(EST_GLOB))
    sco_files = sorted(glob.glob(SCORES_GLOB))

    assert len(ref_files) == len(est_files) == len(sco_files) > 0

    # Unit tests
    for metric in [mir_eval.separation.bss_eval_sources,
                   mir_eval.separation.bss_eval_sources_framewise,
                   mir_eval.separation.bss_eval_images,
                   mir_eval.separation.bss_eval_images_framewise]:
        yield (__unit_test_empty_input, metric)
        yield (__unit_test_silent_input, metric)
        yield (__unit_test_incompatible_shapes, metric)
        yield (__unit_test_too_many_sources, metric)
        yield (__unit_test_too_many_dimensions, metric)
    for metric in [mir_eval.separation.bss_eval_sources,
                   mir_eval.separation.bss_eval_images]:
        yield (__unit_test_default_permutation, metric)
    for metric in [mir_eval.separation.bss_eval_sources_framewise,
                   mir_eval.separation.bss_eval_images_framewise]:
        yield (__unit_test_framewise_small_window, metric)
        yield (__unit_test_partial_silence, metric)
    # Regression tests
    for ref_f, est_f, sco_f in zip(ref_files, est_files, sco_files):
        with open(sco_f, 'r') as f:
            expected_results = json.load(f)
            expected_sources = expected_results['Sources']
            expected_frames = expected_results['Framewise']
            expected_images = expected_results['Images']
            expected_image_frames = expected_results['Images Framewise']
        # Load in example source separation data
        ref_sources = __load_and_stack_wavs(ref_f)
        est_sources = __load_and_stack_wavs(est_f)
        # Test inference for single source passed as single dimensional array
        if ref_sources.shape[0] == 1 and est_sources.shape[0] == 1:
            ref_sources = ref_sources[0]
            est_sources = est_sources[0]

        # Compute scores
        scores = mir_eval.separation.evaluate(
            ref_sources, est_sources,
            window=expected_frames['win'], hop=expected_frames['hop']
        )
        # Compare them
        for metric in scores:
            if 'Sources - ' in metric:
                test_data_name = metric.replace('Sources - ', '')
                # This is a simple hack to make nosetest's messages more useful
                yield (__check_score, sco_f, metric, scores[metric],
                       expected_sources[test_data_name])
            elif 'Sources Frames - ' in metric:
                test_data_name = metric.replace('Sources Frames - ', '')
                # This is a simple hack to make nosetest's messages more useful
                yield (__check_score, sco_f, metric, scores[metric],
                       expected_frames[test_data_name])

        # Compute scores with images
        ref_images = __generate_multichannel(ref_sources,
                                             expected_images['nchan'])
        est_images = __generate_multichannel(est_sources,
                                             expected_images['nchan'],
                                             expected_images['gain'],
                                             expected_images['reverse'])
        image_scores = mir_eval.separation.evaluate(
            ref_images, est_images
        )
        # Compare them
        for metric in image_scores:
            if 'Images - ' in metric:
                test_data_name = metric.replace('Images - ', '')
                # This is a simple hack to make nosetest's messages more useful
                yield (__check_score, sco_f, metric, image_scores[metric],
                       expected_images[test_data_name])

        # Compute scores with images framewise
        ref_images = __generate_multichannel(ref_sources,
                                             expected_image_frames['nchan'])
        est_images = __generate_multichannel(est_sources,
                                             expected_image_frames['nchan'],
                                             expected_image_frames['gain'],
                                             expected_image_frames['reverse'])
        imageframe_scores = mir_eval.separation.evaluate(
            ref_images, est_images,
            window=expected_image_frames['win'],
            hop=expected_image_frames['hop']
        )
        # Compare them
        for metric in imageframe_scores:
            if 'Images Frames - ' in metric:
                test_data_name = metric.replace('Images Frames - ', '')
                # This is a simple hack to make nosetest's messages more useful
                yield (__check_score, sco_f, metric, imageframe_scores[metric],
                       expected_image_frames[test_data_name])

    # Catch a few exceptions in the evaluate function
    image_scores = mir_eval.separation.evaluate(ref_images, est_images)
    # make sure sources is not being evaluated on images
    assert 'Sources - Source to Distortion' not in image_scores



'''
Unit tests for mir_eval.hierarchy
'''

from glob import glob
import re

import warnings
import json

import numpy as np
import scipy.sparse
import mir_eval

from nose.tools import raises


A_TOL = 1e-12


def test_tmeasure_pass():
    # The estimate here gets none of the structure correct.
    ref = [[[0, 30]], [[0, 15], [15, 30]]]
    # convert to arrays
    ref = [np.asarray(_) for _ in ref]

    est = ref[:1]

    def __test(window, frame_size):
        # The estimate should get 0 score here
        scores = mir_eval.hierarchy.tmeasure(ref, est,
                                             window=window,
                                             frame_size=frame_size)

        for k in scores:
            assert k == 0.0

        # The reference should get a perfect score here
        scores = mir_eval.hierarchy.tmeasure(ref, ref,
                                             window=window,
                                             frame_size=frame_size)

        for k in scores:
            assert k == 1.0

    for window in [5, 10, 15, 30, 90, None]:
        for frame_size in [0.1, 0.5, 1.0]:
            yield __test, window, frame_size


def test_tmeasure_warning():

    # Warn if there are missing boundaries from one layer to the next
    ref = [[[0, 5],
            [5, 10]],
           [[0, 10]]]

    ref = [np.asarray(_) for _ in ref]

    warnings.resetwarnings()
    warnings.simplefilter('always')
    with warnings.catch_warnings(record=True) as out:
        mir_eval.hierarchy.tmeasure(ref, ref)

        assert len(out) > 0
        assert out[0].category is UserWarning
        assert ('Segment hierarchy is inconsistent at level 1'
                in str(out[0].message))


def test_tmeasure_fail_span():

    # Does not start at 0
    ref = [[[1, 10]],
           [[1, 5],
            [5, 10]]]

    ref = [np.asarray(_) for _ in ref]

    yield raises(ValueError)(mir_eval.hierarchy.tmeasure), ref, ref

    # Does not end at the right time
    ref = [[[0, 5]],
           [[0, 5],
            [5, 6]]]
    ref = [np.asarray(_) for _ in ref]

    yield raises(ValueError)(mir_eval.hierarchy.tmeasure), ref, ref

    # Two annotaions of different shape
    ref = [[[0, 10]],
           [[0, 5],
            [5, 10]]]
    ref = [np.asarray(_) for _ in ref]

    est = [[[0, 15]],
           [[0, 5],
            [5, 15]]]
    est = [np.asarray(_) for _ in est]

    yield raises(ValueError)(mir_eval.hierarchy.tmeasure), ref, est


def test_tmeasure_fail_frame_size():
    ref = [[[0, 60]],
           [[0, 30],
            [30, 60]]]
    ref = [np.asarray(_) for _ in ref]

    @raises(ValueError)
    def __test(window, frame_size):
        mir_eval.hierarchy.tmeasure(ref, ref,
                                    window=window,
                                    frame_size=frame_size)

    for window in [None, 15, 30]:
        for frame_size in [-1, 0]:
            yield __test, window, frame_size
        if window is not None:
            yield __test, window, 2 * window


def test_lmeasure_pass():

    # The estimate here gets none of the structure correct.
    ref = [[[0, 30]], [[0, 15], [15, 30]]]
    ref_lab = [['A'], ['a', 'b']]

    # convert to arrays
    ref = [np.asarray(_) for _ in ref]

    est = ref[:1]
    est_lab = ref_lab[:1]

    def __test(frame_size):
        # The estimate should get 0 score here
        scores = mir_eval.hierarchy.lmeasure(ref, ref_lab, est, est_lab,
                                             frame_size=frame_size)

        for k in scores:
            assert k == 0.0

        # The reference should get a perfect score here
        scores = mir_eval.hierarchy.lmeasure(ref, ref_lab, ref, ref_lab,
                                             frame_size=frame_size)

        for k in scores:
            assert k == 1.0

    for frame_size in [0.1, 0.5, 1.0]:
        yield __test, frame_size


def test_lmeasure_warning():

    # Warn if there are missing boundaries from one layer to the next
    ref = [[[0, 5],
            [5, 10]],
           [[0, 10]]]

    ref = [np.asarray(_) for _ in ref]
    ref_lab = [['a', 'b'], ['A']]

    warnings.resetwarnings()
    warnings.simplefilter('always')
    with warnings.catch_warnings(record=True) as out:
        mir_eval.hierarchy.lmeasure(ref, ref_lab, ref, ref_lab)

        assert len(out) > 0
        assert out[0].category is UserWarning
        assert ('Segment hierarchy is inconsistent at level 1'
                in str(out[0].message))


def test_lmeasure_fail_span():

    # Does not start at 0
    ref = [[[1, 10]],
           [[1, 5],
            [5, 10]]]

    ref_lab = [['A'], ['a', 'b']]

    ref = [np.asarray(_) for _ in ref]

    yield (raises(ValueError)(mir_eval.hierarchy.lmeasure),
           ref, ref_lab, ref, ref_lab)

    # Does not end at the right time
    ref = [[[0, 5]],
           [[0, 5],
            [5, 6]]]
    ref = [np.asarray(_) for _ in ref]

    yield (raises(ValueError)(mir_eval.hierarchy.lmeasure),
           ref, ref_lab, ref, ref_lab)

    # Two annotations of different shape
    ref = [[[0, 10]],
           [[0, 5],
            [5, 10]]]
    ref = [np.asarray(_) for _ in ref]

    est = [[[0, 15]],
           [[0, 5],
            [5, 15]]]
    est = [np.asarray(_) for _ in est]

    yield (raises(ValueError)(mir_eval.hierarchy.lmeasure),
           ref, ref_lab, est, ref_lab)


def test_lmeasure_fail_frame_size():
    ref = [[[0, 60]],
           [[0, 30],
            [30, 60]]]
    ref = [np.asarray(_) for _ in ref]
    ref_lab = [['A'], ['a', 'b']]

    @raises(ValueError)
    def __test(frame_size):
        mir_eval.hierarchy.lmeasure(ref, ref_lab, ref, ref_lab,
                                    frame_size=frame_size)

    for frame_size in [-1, 0]:
        yield __test, frame_size


def test_hierarchy_regression():

    ref_files = sorted(glob('data/hierarchy/ref*.lab'))
    est_files = sorted(glob('data/hierarchy/est*.lab'))
    out_files = sorted(glob('data/hierarchy/output*.json'))

    ref_hier = [mir_eval.io.load_labeled_intervals(_) for _ in ref_files]
    est_hier = [mir_eval.io.load_labeled_intervals(_) for _ in est_files]

    ref_ints = [seg[0] for seg in ref_hier]
    ref_labs = [seg[1] for seg in ref_hier]
    est_ints = [seg[0] for seg in est_hier]
    est_labs = [seg[1] for seg in est_hier]

    def __test(w, ref_i, ref_l, est_i, est_l, target):
        outputs = mir_eval.hierarchy.evaluate(ref_i, ref_l,
                                              est_i, est_l,
                                              window=w)

        for key in target:
            assert np.allclose(target[key], outputs[key], atol=A_TOL)

    for out in out_files:
        with open(out, 'r') as fdesc:
            target = json.load(fdesc)

        # Extract the window parameter
        window = float(re.match('.*output_w=(\d+).json$', out).groups()[0])
        yield __test, window, ref_ints, ref_labs, est_ints, est_labs, target


def test_count_inversions():

    # inversion count = |{(i, j) : a[i] >= b[j]}|
    a = [2, 4, 6]
    b = [1, 2, 3, 4]

    # All inversions (a, b) are:
    # (2, 1), (2, 2)
    # (4, 1), (4, 2), (4, 3), (4, 4)
    # (6, 1), (6, 2), (6, 3), (6, 4)

    assert mir_eval.hierarchy._count_inversions(a, b) == 10

    # All inversions (b, a) are:
    # (2, 2)
    # (3, 2)
    # (4, 2), (4, 4)

    assert mir_eval.hierarchy._count_inversions(b, a) == 4

    # And test with repetitions
    a = [2, 2, 4]
    b = [1, 2, 4, 4]
    # counts: (a, b)
    # (2, 1), (2, 2)
    # (2, 1), (2, 2)
    # (4, 1), (4, 2), (4, 4), (4, 4)

    assert mir_eval.hierarchy._count_inversions(a, b) == 8

    # count: (b, a)
    # (2, 2), (2, 2)
    # (4, 2), (4, 2), (4, 4)
    # (4, 2), (4, 2), (4, 4)

    assert mir_eval.hierarchy._count_inversions(b, a) == 8


def test_meet():

    frame_size = 1
    int_hier = [np.array([[0, 10]]),
                np.array([[0, 6], [6, 10]]),
                np.array([[0, 2], [2, 4], [4, 6], [6, 8], [8, 10]])]

    lab_hier = [['X'],
                ['A', 'B'],
                ['a', 'b', 'a', 'c', 'b']]

    # Target output
    meet_truth = np.asarray([
        [3, 3, 2, 2, 3, 3, 1, 1, 1, 1],   # (XAa)
        [3, 3, 2, 2, 3, 3, 1, 1, 1, 1],   # (XAa)
        [2, 2, 3, 3, 2, 2, 1, 1, 3, 3],   # (XAb)
        [2, 2, 3, 3, 2, 2, 1, 1, 3, 3],   # (XAb)
        [3, 3, 2, 2, 3, 3, 1, 1, 1, 1],   # (XAa)
        [3, 3, 2, 2, 3, 3, 1, 1, 1, 1],   # (XAa)
        [1, 1, 1, 1, 1, 1, 3, 3, 2, 2],   # (XBc)
        [1, 1, 1, 1, 1, 1, 3, 3, 2, 2],   # (XBc)
        [1, 1, 3, 3, 1, 1, 2, 2, 3, 3],   # (XBb)
        [1, 1, 3, 3, 1, 1, 2, 2, 3, 3],   # (XBb)
    ])
    meet = mir_eval.hierarchy._meet(int_hier, lab_hier, frame_size)

    # Is it the right type?
    assert isinstance(meet, scipy.sparse.csr_matrix)
    meet = meet.toarray()

    # Does it have the right shape?
    assert meet.shape == (10, 10)

    # Does it have the right value?
    assert np.all(meet == meet_truth)


def test_compare_frame_rankings():

    # number of pairs (i, j)
    # where ref[i] < ref[j] and est[i] >= est[j]

    ref = np.asarray([1, 2, 3, 3])
    # ref pairs (transitive)
    # (1, 2), (1, 3), (1, 3), (2, 3), (2, 3)
    # ref pairs (non-transitive)
    # (1, 2), (2, 3), (2, 3)

    # Just count the normalizers
    # No self-inversions are possible from ref to itself
    inv, norm = mir_eval.hierarchy._compare_frame_rankings(ref, ref,
                                                           transitive=True)
    assert inv == 0
    assert norm == 5.0

    inv, norm = mir_eval.hierarchy._compare_frame_rankings(ref, ref,
                                                           transitive=False)
    assert inv == 0
    assert norm == 3.0

    est = np.asarray([1, 2, 1, 3])
    # In the transitive case, we lose two pairs
    # (1, 3) and (2, 2) -> (1, 1), (2, 1)
    inv, norm = mir_eval.hierarchy._compare_frame_rankings(ref, est,
                                                           transitive=True)
    assert inv == 2
    assert norm == 5.0

    # In the non-transitive case, we only lose one pair
    # because (1,3) was not counted
    inv, norm = mir_eval.hierarchy._compare_frame_rankings(ref, est,
                                                           transitive=False)
    assert inv == 1
    assert norm == 3.0

    # Do an all-zeros test
    ref = np.asarray([1, 1, 1, 1])
    inv, norm = mir_eval.hierarchy._compare_frame_rankings(ref, ref,
                                                           transitive=True)
    assert inv == 0
    assert norm == 0.0



"""
Unit tests for mir_eval.multipitch
"""

import numpy as np
import json
import mir_eval
import glob
import warnings
import nose.tools

A_TOL = 1e-12

# Path to the fixture files
REF_GLOB = 'data/multipitch/ref*.txt'
EST_GLOB = 'data/multipitch/est*.txt'
SCORES_GLOB = 'data/multipitch/output*.json'


def __frequencies_equal(freqs_a, freqs_b):
    if len(freqs_a) != len(freqs_b):
        return False
    else:
        equal = True
        for freq_a, freq_b in zip(freqs_a, freqs_b):
            if freq_a.size != freq_b.size:
                return False
            equal = equal and np.allclose(freq_a, freq_b, atol=A_TOL)
        return equal


def __scores_equal(score_a, score_b):
    keys_a = set(list(score_a.keys()))
    keys_b = set(list(score_b.keys()))
    if keys_a != keys_b:
        return False
    else:
        equal = True
        for k in keys_a:
            value_a = score_a[k]
            value_b = score_b[k]
            equal = equal and np.allclose(value_a, value_b, atol=A_TOL)
    return equal


def test_resample_multif0():
    times = np.array([0.00, 0.01, 0.02, 0.03])
    empty_times = np.array([])
    freqs = [
        np.array([200.]),
        np.array([]),
        np.array([300., 400., 500.]),
        np.array([300., 500.])
    ]
    empty_freqs = []
    target_times1 = times
    target_times2 = np.array([0.001, 0.002, 0.01, 0.029, 0.05])
    target_times3 = empty_times

    expected_freqs1 = freqs
    expected_freqs2 = [
        np.array([200.]),
        np.array([200.]),
        np.array([]),
        np.array([300., 500.]),
        np.array([])
    ]
    expected_freqs3 = empty_freqs
    expected_freqs4 = [np.array([])]*4

    actual_freqs1 = mir_eval.multipitch.resample_multipitch(
        times, freqs, target_times1)
    actual_freqs2 = mir_eval.multipitch.resample_multipitch(
        times, freqs, target_times2)
    actual_freqs3 = mir_eval.multipitch.resample_multipitch(
        times, freqs, target_times3)
    actual_freqs4 = mir_eval.multipitch.resample_multipitch(
        empty_times, empty_freqs, target_times1)

    assert __frequencies_equal(actual_freqs1, expected_freqs1)
    assert __frequencies_equal(actual_freqs2, expected_freqs2)
    assert __frequencies_equal(actual_freqs3, expected_freqs3)
    assert __frequencies_equal(actual_freqs4, expected_freqs4)


def test_frequencies_to_midi():
    frequencies = [
        np.array([440.]),
        np.array([]),
        np.array([220., 660., 512.]),
        np.array([300., 512.])
    ]
    expected = [
        np.array([69.]),
        np.array([]),
        np.array([57., 76.01955000865388, 71.623683437704088]),
        np.array([62.369507723654657, 71.623683437704088])
    ]
    actual = mir_eval.multipitch.frequencies_to_midi(frequencies)
    assert __frequencies_equal(actual, expected)


def test_midi_to_chroma():
    midi_frequencies = [
        np.array([69.]),
        np.array([]),
        np.array([57., 76.01955000865388, 71.623683437704088]),
        np.array([62.369507723654657, 71.623683437704088])
    ]
    expected = [
        np.array([9.]),
        np.array([]),
        np.array([9., 4.01955000865388, 11.623683437704088]),
        np.array([2.3695077236546567, 11.623683437704088])
    ]
    actual = mir_eval.multipitch.midi_to_chroma(midi_frequencies)
    assert __frequencies_equal(actual, expected)


def test_compute_num_freqs():
    frequencies = [
        np.array([256.]),
        np.array([]),
        np.array([362.03867196751236, 128., 512.]),
        np.array([300., 512.])
    ]
    expected = np.array([1, 0, 3, 2])
    actual = mir_eval.multipitch.compute_num_freqs(frequencies)
    assert np.allclose(actual, expected, atol=A_TOL)


def test_compute_num_true_positives():
    ref_freqs = [
        np.array([96., 100.]),
        np.array([]),
        np.array([81.]),
        np.array([102., 84., 108.]),
        np.array([98.745824285950576, 108.])
    ]
    est_freqs = [
        np.array([96.]),
        np.array([]),
        np.array([200., 82.]),
        np.array([102., 84., 108.]),
        np.array([99., 108.])
    ]
    expected = np.array([1, 0, 0, 3, 2])
    actual = mir_eval.multipitch.compute_num_true_positives(
        ref_freqs, est_freqs)
    assert np.allclose(actual, expected, atol=A_TOL)

    ref_freqs_chroma = [
        np.array([0., 1.5]),
        np.array([]),
        np.array([2.]),
        np.array([5.1, 6., 11.]),
        np.array([11.9, 11.9])
    ]
    est_freqs_chroma = [
        np.array([0.]),
        np.array([]),
        np.array([5., 2.6]),
        np.array([5.1, 6., 11.]),
        np.array([0.2, 11.5])
    ]
    expected = np.array([1, 0, 0, 3, 2])
    actual = mir_eval.multipitch.compute_num_true_positives(
        ref_freqs_chroma, est_freqs_chroma, chroma=True)
    assert np.allclose(actual, expected, atol=A_TOL)


def test_accuracy_metrics():
    true_positives = np.array([1, 0, 0, 3, 2])
    n_ref = np.array([2, 0, 1, 3, 2])
    n_est = np.array([1, 0, 2, 3, 2])

    expected_precision = 0.75
    expected_recall = 0.75
    expected_accuracy = 0.6

    (actual_precision,
     actual_recall,
     actual_accuarcy) = mir_eval.multipitch.compute_accuracy(
         true_positives, n_ref, n_est)

    assert np.allclose(actual_precision, expected_precision, atol=A_TOL)
    assert np.allclose(actual_recall, expected_recall, atol=A_TOL)
    assert np.allclose(actual_accuarcy, expected_accuracy, atol=A_TOL)


def test_error_score_metrics():
    true_positives = np.array([1, 0, 0, 3, 2])
    n_ref = np.array([2, 0, 1, 3, 2])
    n_est = np.array([1, 0, 2, 3, 2])

    expected_esub = 0.125
    expected_emiss = 0.125
    expected_efa = 0.125
    expected_etot = 0.375

    (actual_esub,
     actual_emiss,
     actual_efa,
     actual_etot) = mir_eval.multipitch.compute_err_score(
         true_positives, n_ref, n_est)

    assert np.allclose(actual_esub, expected_esub, atol=A_TOL)
    assert np.allclose(actual_emiss, expected_emiss, atol=A_TOL)
    assert np.allclose(actual_efa, expected_efa, atol=A_TOL)
    assert np.allclose(actual_etot, expected_etot, atol=A_TOL)


def unit_test_metrics():
    empty_array = np.array([])
    ref_time = np.array([0.0, 0.1])
    ref_freqs = [np.array([201.]), np.array([])]
    est_time = np.array([0.0, 0.1])
    est_freqs = [np.array([200.]), np.array([])]

    # ref sizes unequal
    nose.tools.assert_raises(
        ValueError, mir_eval.multipitch.metrics,
        np.array([0.0]), ref_freqs, est_time, est_freqs)

    # est sizes unequal
    nose.tools.assert_raises(
        ValueError, mir_eval.multipitch.metrics,
        ref_time, ref_freqs, np.array([0.0]), est_freqs)

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter('always')
        # Test for warnings on empty values
        actual_score = mir_eval.multipitch.metrics(
            ref_time, [empty_array, empty_array],
            est_time, [empty_array, empty_array])
        assert len(w) == 6
        assert issubclass(w[-1].category, UserWarning)
        assert str(w[-1].message) == "Reference frequencies are all empty."

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter('always')
        # Test for warnings on empty values
        # test all inputs empty
        mir_eval.multipitch.metrics(empty_array, [], empty_array, [])
        assert len(w) == 10
        assert issubclass(w[-1].category, UserWarning)
        assert str(w[-1].message) == "Reference frequencies are all empty."

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter('always')
        # reference empty
        mir_eval.multipitch.metrics(empty_array, [], est_time, est_freqs)
        assert len(w) == 9
        assert issubclass(w[-1].category, UserWarning)
        assert str(w[-1].message) == "Reference frequencies are all empty."

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter('always')
        # estimate empty
        mir_eval.multipitch.metrics(ref_time, ref_freqs, empty_array, [])
        assert len(w) == 5
        assert issubclass(w[-1].category, UserWarning)
        assert str(w[-1].message) == "Estimate frequencies are all empty."

        expected_score = (0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                          0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0)
        assert np.allclose(actual_score, expected_score)

    # test perfect estimate
    ref_time = np.array([0.0, 0.1, 0.2])
    ref_freqs = [np.array([201.]), np.array([]), np.array([300.5, 87.1])]
    actual_score = mir_eval.multipitch.metrics(
        ref_time, ref_freqs, ref_time, ref_freqs)

    expected_score = (1.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0,
                      1.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0)
    assert np.allclose(actual_score, expected_score)


def regression_test_evaluate():
    ref_files = sorted(glob.glob(REF_GLOB))
    est_files = sorted(glob.glob(EST_GLOB))
    score_files = sorted(glob.glob(SCORES_GLOB))

    assert len(ref_files) == len(est_files) == len(score_files) > 0

    for ref_f, est_f, score_f in zip(ref_files, est_files, score_files):
        with open(score_f, 'r') as f_handle:
            expected_score = json.load(f_handle)

        ref_times, ref_freqs = mir_eval.io.load_ragged_time_series(ref_f)
        est_times, est_freqs = mir_eval.io.load_ragged_time_series(est_f)

        actual_score = mir_eval.multipitch.evaluate(
            ref_times, ref_freqs, est_times, est_freqs)

        assert __scores_equal(actual_score, expected_score)



"""
Transcription evaluation, as defined in :mod:`mir_eval.transcription`, does not
take into account the velocities of reference and estimated notes. This
submodule implements a variant of
:func:`mir_eval.transcription.precision_recall_f1_overlap` which
additionally considers note velocity when determining whether a note is
correctly transcribed. This is done by defining a new function
:func:`mir_eval.transcription_velocity.match_notes` which first calls
:func:`mir_eval.transcription.match_notes` to get a note matching based on
onset, offset, and pitch. Then, we follow the evaluation procedure described in
[#hawthorne2018onsets]_ to test whether an estimated note should be considered
correct:

1. Reference velocities are re-scaled to the range [0, 1].

2. A linear regression is performed to estimate global scale and offset
   parameters which minimize the L2 distance between matched estimated and
   (rescaled) reference notes.

3. The scale and offset parameters are used to rescale estimated
   velocities.

4. An estimated/reference note pair which has been matched according to the
   onset, offset, and pitch is further only considered correct if the rescaled
   velocities are within a predefined threshold, defaulting to 0.1.

:func:`mir_eval.transcription_velocity.match_notes` is used to define a new
variant :func:`mir_eval.transcription_velocity.precision_recall_f1_overlap`
which considers velocity.

Conventions
-----------

This submodule follows the conventions of :mod:`mir_eval.transcription` and
additionally requires velocities to be provided as MIDI velocities in the range
[0, 127].

Metrics
-------

* :func:`mir_eval.transcription_velocity.precision_recall_f1_overlap`: The
  precision, recall, F-measure, and Average Overlap Ratio of the note
  transcription, where an estimated note is considered correct if its pitch,
  onset, velocity and (optionally) offset are sufficiently close to a reference
  note.

References
----------
  .. [#hawthorne2018onsets] Curtis Hawthorne, Erich Elsen, Jialin Song, Adam
      Roberts, Ian Simon, Colin Raffel, Jesse Engel, Sageev Oore, and Douglas
      Eck, "Onsets and Frames: Dual-Objective Piano Transcription", Proceedings
      of the 19th International Society for Music Information Retrieval
      Conference, 2018.
"""

import collections
import numpy as np
from . import transcription
from . import util


def validate(ref_intervals, ref_pitches, ref_velocities, est_intervals,
             est_pitches, est_velocities):
    """Checks that the input annotations have valid time intervals, pitches,
    and velocities, and throws helpful errors if not.

    Parameters
    ----------
    ref_intervals : np.ndarray, shape=(n,2)
        Array of reference notes time intervals (onset and offset times)
    ref_pitches : np.ndarray, shape=(n,)
        Array of reference pitch values in Hertz
    ref_velocities : np.ndarray, shape=(n,)
        Array of MIDI velocities (i.e. between 0 and 127) of reference notes
    est_intervals : np.ndarray, shape=(m,2)
        Array of estimated notes time intervals (onset and offset times)
    est_pitches : np.ndarray, shape=(m,)
        Array of estimated pitch values in Hertz
    est_velocities : np.ndarray, shape=(m,)
        Array of MIDI velocities (i.e. between 0 and 127) of estimated notes
    """
    transcription.validate(ref_intervals, ref_pitches, est_intervals,
                           est_pitches)
    # Check that velocities have the same length as intervals/pitches
    if not ref_velocities.shape[0] == ref_pitches.shape[0]:
        raise ValueError('Reference velocities must have the same length as '
                         'pitches and intervals.')
    if not est_velocities.shape[0] == est_pitches.shape[0]:
        raise ValueError('Estimated velocities must have the same length as '
                         'pitches and intervals.')
    # Check that the velocities are positive
    if ref_velocities.size > 0 and np.min(ref_velocities) < 0:
        raise ValueError('Reference velocities must be positive.')
    if est_velocities.size > 0 and np.min(est_velocities) < 0:
        raise ValueError('Estimated velocities must be positive.')


def match_notes(
        ref_intervals, ref_pitches, ref_velocities, est_intervals, est_pitches,
        est_velocities, onset_tolerance=0.05, pitch_tolerance=50.0,
        offset_ratio=0.2, offset_min_tolerance=0.05, strict=False,
        velocity_tolerance=0.1):
    """Match notes, taking note velocity into consideration.

    This function first calls :func:`mir_eval.transcription.match_notes` to
    match notes according to the supplied intervals, pitches, onset, offset,
    and pitch tolerances. The velocities of the matched notes are then used to
    estimate a slope and intercept which can rescale the estimated velocities
    so that they are as close as possible (in L2 sense) to their matched
    reference velocities. Velocities are then normalized to the range [0, 1]. A
    estimated note is then further only considered correct if its velocity is
    within ``velocity_tolerance`` of its matched (according to pitch and
    timing) reference note.

    Parameters
    ----------
    ref_intervals : np.ndarray, shape=(n,2)
        Array of reference notes time intervals (onset and offset times)
    ref_pitches : np.ndarray, shape=(n,)
        Array of reference pitch values in Hertz
    ref_velocities : np.ndarray, shape=(n,)
        Array of MIDI velocities (i.e. between 0 and 127) of reference notes
    est_intervals : np.ndarray, shape=(m,2)
        Array of estimated notes time intervals (onset and offset times)
    est_pitches : np.ndarray, shape=(m,)
        Array of estimated pitch values in Hertz
    est_velocities : np.ndarray, shape=(m,)
        Array of MIDI velocities (i.e. between 0 and 127) of estimated notes
    onset_tolerance : float > 0
        The tolerance for an estimated note's onset deviating from the
        reference note's onset, in seconds. Default is 0.05 (50 ms).
    pitch_tolerance : float > 0
        The tolerance for an estimated note's pitch deviating from the
        reference note's pitch, in cents. Default is 50.0 (50 cents).
    offset_ratio : float > 0 or None
        The ratio of the reference note's duration used to define the
        offset_tolerance. Default is 0.2 (20%), meaning the
        ``offset_tolerance`` will equal the ``ref_duration * 0.2``, or 0.05 (50
        ms), whichever is greater. If ``offset_ratio`` is set to ``None``,
        offsets are ignored in the matching.
    offset_min_tolerance : float > 0
        The minimum tolerance for offset matching. See offset_ratio description
        for an explanation of how the offset tolerance is determined. Note:
        this parameter only influences the results if ``offset_ratio`` is not
        ``None``.
    strict : bool
        If ``strict=False`` (the default), threshold checks for onset, offset,
        and pitch matching are performed using ``<=`` (less than or equal). If
        ``strict=True``, the threshold checks are performed using ``<`` (less
        than).
    velocity_tolerance : float > 0
        Estimated notes are considered correct if, after rescaling and
        normalization to [0, 1], they are within ``velocity_tolerance`` of a
        matched reference note.

    Returns
    -------
    matching : list of tuples
        A list of matched reference and estimated notes.
        ``matching[i] == (i, j)`` where reference note ``i`` matches estimated
        note ``j``.
    """
    # Compute note matching as usual using standard transcription function
    matching = transcription.match_notes(
        ref_intervals, ref_pitches, est_intervals, est_pitches,
        onset_tolerance, pitch_tolerance, offset_ratio, offset_min_tolerance,
        strict)

    # Rescale reference velocities to the range [0, 1]
    min_velocity, max_velocity = np.min(ref_velocities), np.max(ref_velocities)
    # Make the smallest possible range 1 to avoid divide by zero
    velocity_range = max(1, max_velocity - min_velocity)
    ref_velocities = (ref_velocities - min_velocity)/float(velocity_range)

    # Convert matching list-of-tuples to array for fancy indexing
    matching = np.array(matching)
    # When there is no matching, return an empty list
    if matching.size == 0:
        return []
    # Grab velocities for matched notes
    ref_matched_velocities = ref_velocities[matching[:, 0]]
    est_matched_velocities = est_velocities[matching[:, 1]]
    # Find slope and intercept of line which produces best least-squares fit
    # between matched est and ref velocities
    slope, intercept = np.linalg.lstsq(
        np.vstack([est_matched_velocities,
                   np.ones(len(est_matched_velocities))]).T,
        ref_matched_velocities)[0]
    # Re-scale est velocities to match ref
    est_matched_velocities = slope*est_matched_velocities + intercept
    # Compute the absolute error of (rescaled) estimated velocities vs.
    # normalized reference velocities. Error will be in [0, 1]
    velocity_diff = np.abs(est_matched_velocities - ref_matched_velocities)
    # Check whether each error is within the provided tolerance
    velocity_within_tolerance = (velocity_diff < velocity_tolerance)
    # Only keep matches whose velocity was within the provided tolerance
    matching = matching[velocity_within_tolerance]
    # Convert back to list-of-tuple format
    matching = [tuple(_) for _ in matching]

    return matching


def precision_recall_f1_overlap(
        ref_intervals, ref_pitches, ref_velocities, est_intervals, est_pitches,
        est_velocities, onset_tolerance=0.05, pitch_tolerance=50.0,
        offset_ratio=0.2, offset_min_tolerance=0.05, strict=False,
        velocity_tolerance=0.1, beta=1.0):
    """Compute the Precision, Recall and F-measure of correct vs incorrectly
    transcribed notes, and the Average Overlap Ratio for correctly transcribed
    notes (see :func:`mir_eval.transcription.average_overlap_ratio`).
    "Correctness" is determined based on note onset, velocity, pitch and
    (optionally) offset. An estimated note is considered correct if

    1. Its onset is within ``onset_tolerance`` (default +-50ms) of a
       reference note
    2. Its pitch (F0) is within +/- ``pitch_tolerance`` (default one
       quarter tone, 50 cents) of the corresponding reference note
    3. Its velocity, after normalizing reference velocities to the range
       [0, 1] and globally rescaling estimated velocities to minimize L2
       distance between matched reference notes, is within
       ``velocity_tolerance`` (default 0.1) the corresponding reference note
    4. If ``offset_ratio`` is ``None``, note offsets are ignored in the
       comparison. Otherwise, on top of the above requirements, a correct
       returned note is required to have an offset value within
       `offset_ratio`` (default 20%) of the reference note's duration around
       the reference note's offset, or within ``offset_min_tolerance``
       (default 50 ms), whichever is larger.

    Parameters
    ----------
    ref_intervals : np.ndarray, shape=(n,2)
        Array of reference notes time intervals (onset and offset times)
    ref_pitches : np.ndarray, shape=(n,)
        Array of reference pitch values in Hertz
    ref_velocities : np.ndarray, shape=(n,)
        Array of MIDI velocities (i.e. between 0 and 127) of reference notes
    est_intervals : np.ndarray, shape=(m,2)
        Array of estimated notes time intervals (onset and offset times)
    est_pitches : np.ndarray, shape=(m,)
        Array of estimated pitch values in Hertz
    est_velocities : np.ndarray, shape=(n,)
        Array of MIDI velocities (i.e. between 0 and 127) of estimated notes
    onset_tolerance : float > 0
        The tolerance for an estimated note's onset deviating from the
        reference note's onset, in seconds. Default is 0.05 (50 ms).
    pitch_tolerance : float > 0
        The tolerance for an estimated note's pitch deviating from the
        reference note's pitch, in cents. Default is 50.0 (50 cents).
    offset_ratio : float > 0 or None
        The ratio of the reference note's duration used to define the
        offset_tolerance. Default is 0.2 (20%), meaning the
        ``offset_tolerance`` will equal the ``ref_duration * 0.2``, or
        ``offset_min_tolerance`` (0.05 by default, i.e. 50 ms), whichever is
        greater. If ``offset_ratio`` is set to ``None``, offsets are ignored in
        the evaluation.
    offset_min_tolerance : float > 0
        The minimum tolerance for offset matching. See ``offset_ratio``
        description for an explanation of how the offset tolerance is
        determined. Note: this parameter only influences the results if
        ``offset_ratio`` is not ``None``.
    strict : bool
        If ``strict=False`` (the default), threshold checks for onset, offset,
        and pitch matching are performed using ``<=`` (less than or equal). If
        ``strict=True``, the threshold checks are performed using ``<`` (less
        than).
    velocity_tolerance : float > 0
        Estimated notes are considered correct if, after rescaling and
        normalization to [0, 1], they are within ``velocity_tolerance`` of a
        matched reference note.
    beta : float > 0
        Weighting factor for f-measure (default value = 1.0).

    Returns
    -------
    precision : float
        The computed precision score
    recall : float
        The computed recall score
    f_measure : float
        The computed F-measure score
    avg_overlap_ratio : float
        The computed Average Overlap Ratio score
    """
    validate(ref_intervals, ref_pitches, ref_velocities, est_intervals,
             est_pitches, est_velocities)
    # When reference notes are empty, metrics are undefined, return 0's
    if len(ref_pitches) == 0 or len(est_pitches) == 0:
        return 0., 0., 0., 0.

    matching = match_notes(
        ref_intervals, ref_pitches, ref_velocities, est_intervals, est_pitches,
        est_velocities, onset_tolerance, pitch_tolerance, offset_ratio,
        offset_min_tolerance, strict, velocity_tolerance)

    precision = float(len(matching))/len(est_pitches)
    recall = float(len(matching))/len(ref_pitches)
    f_measure = util.f_measure(precision, recall, beta=beta)

    avg_overlap_ratio = transcription.average_overlap_ratio(
        ref_intervals, est_intervals, matching)

    return precision, recall, f_measure, avg_overlap_ratio


def evaluate(ref_intervals, ref_pitches, ref_velocities, est_intervals,
             est_pitches, est_velocities, **kwargs):
    """Compute all metrics for the given reference and estimated annotations.

    Parameters
    ----------
    ref_intervals : np.ndarray, shape=(n,2)
        Array of reference notes time intervals (onset and offset times)
    ref_pitches : np.ndarray, shape=(n,)
        Array of reference pitch values in Hertz
    ref_velocities : np.ndarray, shape=(n,)
        Array of MIDI velocities (i.e. between 0 and 127) of reference notes
    est_intervals : np.ndarray, shape=(m,2)
        Array of estimated notes time intervals (onset and offset times)
    est_pitches : np.ndarray, shape=(m,)
        Array of estimated pitch values in Hertz
    est_velocities : np.ndarray, shape=(n,)
        Array of MIDI velocities (i.e. between 0 and 127) of estimated notes
    kwargs
        Additional keyword arguments which will be passed to the
        appropriate metric or preprocessing functions.

    Returns
    -------
    scores : dict
        Dictionary of scores, where the key is the metric name (str) and
        the value is the (float) score achieved.
    """
    # Compute all the metrics
    scores = collections.OrderedDict()

    # Precision, recall and f-measure taking note offsets into account
    kwargs.setdefault('offset_ratio', 0.2)
    if kwargs['offset_ratio'] is not None:
        (scores['Precision'],
         scores['Recall'],
         scores['F-measure'],
         scores['Average_Overlap_Ratio']) = util.filter_kwargs(
             precision_recall_f1_overlap, ref_intervals, ref_pitches,
             ref_velocities, est_intervals, est_pitches, est_velocities,
             **kwargs)

    # Precision, recall and f-measure NOT taking note offsets into account
    kwargs['offset_ratio'] = None
    (scores['Precision_no_offset'],
     scores['Recall_no_offset'],
     scores['F-measure_no_offset'],
     scores['Average_Overlap_Ratio_no_offset']) = util.filter_kwargs(
         precision_recall_f1_overlap, ref_intervals, ref_pitches,
         ref_velocities, est_intervals, est_pitches, est_velocities, **kwargs)

    return scores



'''
Unit tests for mir_eval.onset
'''

import numpy as np
import json
import mir_eval
import glob
import warnings
import nose.tools

A_TOL = 1e-12

# Path to the fixture files
REF_GLOB = 'data/onset/ref*.txt'
EST_GLOB = 'data/onset/est*.txt'
SCORES_GLOB = 'data/onset/output*.json'


def __unit_test_onset_function(metric):
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter('always')
        # First, test for a warning on empty onsets
        metric(np.array([]), np.arange(10))
        assert len(w) == 1
        assert issubclass(w[-1].category, UserWarning)
        assert str(w[-1].message) == "Reference onsets are empty."
        metric(np.arange(10), np.array([]))
        assert len(w) == 2
        assert issubclass(w[-1].category, UserWarning)
        assert str(w[-1].message) == "Estimated onsets are empty."
        # And that the metric is 0
        assert np.allclose(metric(np.array([]), np.array([])), 0)

    # Now test validation function - onsets must be 1d ndarray
    onsets = np.array([[1., 2.]])
    nose.tools.assert_raises(ValueError, metric, onsets, onsets)
    # onsets must be in seconds (so not huge)
    onsets = np.array([1e10, 1e11])
    nose.tools.assert_raises(ValueError, metric, onsets, onsets)
    # onsets must be sorted
    onsets = np.array([2., 1.])
    nose.tools.assert_raises(ValueError, metric, onsets, onsets)

    # Valid onsets which are the same produce a score of 1 for all metrics
    onsets = np.arange(10, dtype=np.float)
    assert np.allclose(metric(onsets, onsets), 1)


def __check_score(sco_f, metric, score, expected_score):
    assert np.allclose(score, expected_score, atol=A_TOL)


def test_onset_functions():
    # Load in all files in the same order
    ref_files = sorted(glob.glob(REF_GLOB))
    est_files = sorted(glob.glob(EST_GLOB))
    sco_files = sorted(glob.glob(SCORES_GLOB))

    assert len(ref_files) == len(est_files) == len(sco_files) > 0

    # Unit tests
    for metric in [mir_eval.onset.f_measure]:
        yield (__unit_test_onset_function, metric)
    # Regression tests
    for ref_f, est_f, sco_f in zip(ref_files, est_files, sco_files):
        with open(sco_f, 'r') as f:
            expected_scores = json.load(f)
        # Load in an example onset annotation
        reference_onsets = mir_eval.io.load_events(ref_f)
        # Load in an example onset tracker output
        estimated_onsets = mir_eval.io.load_events(est_f)
        # Compute scores
        scores = mir_eval.onset.evaluate(reference_onsets, estimated_onsets)
        # Compare them
        for metric in scores:
            # This is a simple hack to make nosetest's messages more useful
            yield (__check_score, sco_f, metric, scores[metric],
                   expected_scores[metric])



'''
Methods which sonify annotations for "evaluation by ear".
All functions return a raw signal at the specified sampling rate.
'''

import numpy as np
from numpy.lib.stride_tricks import as_strided
from scipy.interpolate import interp1d

from . import util
from . import chord


def clicks(times, fs, click=None, length=None):
    """Returns a signal with the signal 'click' placed at each specified time

    Parameters
    ----------
    times : np.ndarray
        times to place clicks, in seconds
    fs : int
        desired sampling rate of the output signal
    click : np.ndarray
        click signal, defaults to a 1 kHz blip
    length : int
        desired number of samples in the output signal,
        defaults to ``times.max()*fs + click.shape[0] + 1``

    Returns
    -------
    click_signal : np.ndarray
        Synthesized click signal

    """
    # Create default click signal
    if click is None:
        # 1 kHz tone, 100ms
        click = np.sin(2*np.pi*np.arange(fs*.1)*1000/(1.*fs))
        # Exponential decay
        click *= np.exp(-np.arange(fs*.1)/(fs*.01))
    # Set default length
    if length is None:
        length = int(times.max()*fs + click.shape[0] + 1)

    # Pre-allocate click signal
    click_signal = np.zeros(length)
    # Place clicks
    for time in times:
        # Compute the boundaries of the click
        start = int(time*fs)
        end = start + click.shape[0]
        # Make sure we don't try to output past the end of the signal
        if start >= length:
            break
        if end >= length:
            click_signal[start:] = click[:length - start]
            break
        # Normally, just add a click here
        click_signal[start:end] = click
    return click_signal


def time_frequency(gram, frequencies, times, fs, function=np.sin, length=None,
                   n_dec=1):
    """Reverse synthesis of a time-frequency representation of a signal

    Parameters
    ----------
    gram : np.ndarray
        ``gram[n, m]`` is the magnitude of ``frequencies[n]``
        from ``times[m]`` to ``times[m + 1]``

        Non-positive magnitudes are interpreted as silence.

    frequencies : np.ndarray
        array of size ``gram.shape[0]`` denoting the frequency of
        each row of gram
    times : np.ndarray, shape= ``(gram.shape[1],)`` or ``(gram.shape[1], 2)``
        Either the start time of each column in the gram,
        or the time interval corresponding to each column.
    fs : int
        desired sampling rate of the output signal
    function : function
        function to use to synthesize notes, should be :math:`2\pi`-periodic
    length : int
        desired number of samples in the output signal,
        defaults to ``times[-1]*fs``
    n_dec : int
        the number of decimals used to approximate each sonfied frequency.
        Defaults to 1 decimal place. Higher precision will be slower.

    Returns
    -------
    output : np.ndarray
        synthesized version of the piano roll

    """
    # Default value for length
    if times.ndim == 1:
        # Convert to intervals
        times = util.boundaries_to_intervals(times)

    if length is None:
        length = int(times[-1, 1] * fs)

    times, _ = util.adjust_intervals(times, t_max=length)

    # Truncate times so that the shape matches gram
    n_times = gram.shape[1]
    times = times[:n_times]

    def _fast_synthesize(frequency):
        """A faster way to synthesize a signal.
            Generate one cycle, and simulate arbitrary repetitions
            using array indexing tricks.
        """
        # hack so that we can ensure an integer number of periods and samples
        # rounds frequency to 1st decimal, s.t. 10 * frequency will be an int
        frequency = np.round(frequency, n_dec)

        # Generate 10*frequency periods at this frequency
        # Equivalent to n_samples = int(n_periods * fs / frequency)
        # n_periods = 10*frequency is the smallest integer that guarantees
        # that n_samples will be an integer, since assuming 10*frequency
        # is an integer
        n_samples = int(10.0**n_dec * fs)

        short_signal = function(2.0 * np.pi * np.arange(n_samples) *
                                frequency / fs)

        # Calculate the number of loops we need to fill the duration
        n_repeats = int(np.ceil(length/float(short_signal.shape[0])))

        # Simulate tiling the short buffer by using stride tricks
        long_signal = as_strided(short_signal,
                                 shape=(n_repeats, len(short_signal)),
                                 strides=(0, short_signal.itemsize))

        # Use a flatiter to simulate a long 1D buffer
        return long_signal.flat

    def _const_interpolator(value):
        """Return a function that returns `value`
            no matter the input.
        """
        def __interpolator(x):
            return value
        return __interpolator

    # Threshold the tfgram to remove non-positive values
    gram = np.maximum(gram, 0)

    # Pre-allocate output signal
    output = np.zeros(length)
    time_centers = np.mean(times, axis=1) * float(fs)

    for n, frequency in enumerate(frequencies):
        # Get a waveform of length samples at this frequency
        wave = _fast_synthesize(frequency)

        # Interpolate the values in gram over the time grid
        if len(time_centers) > 1:
            gram_interpolator = interp1d(
                time_centers, gram[n, :],
                kind='linear', bounds_error=False,
                fill_value=(gram[n, 0], gram[n, -1]))
        # If only one time point, create constant interpolator
        else:
            gram_interpolator = _const_interpolator(gram[n, 0])

        # Scale each time interval by the piano roll magnitude
        for m, (start, end) in enumerate((times * fs).astype(int)):
            # Clip the timings to make sure the indices are valid
            start, end = max(start, 0), min(end, length)
            # add to waveform
            output[start:end] += (
                wave[start:end] * gram_interpolator(np.arange(start, end)))

    # Normalize, but only if there's non-zero values
    norm = np.abs(output).max()
    if norm >= np.finfo(output.dtype).tiny:
        output /= norm

    return output


def pitch_contour(times, frequencies, fs, amplitudes=None, function=np.sin,
                  length=None, kind='linear'):
    '''Sonify a pitch contour.

    Parameters
    ----------
    times : np.ndarray
        time indices for each frequency measurement, in seconds

    frequencies : np.ndarray
        frequency measurements, in Hz.
        Non-positive measurements will be interpreted as un-voiced samples.

    fs : int
        desired sampling rate of the output signal

    amplitudes : np.ndarray
        amplitude measurments, nonnegative
        defaults to ``np.ones((length,))``

    function : function
        function to use to synthesize notes, should be :math:`2\pi`-periodic

    length : int
        desired number of samples in the output signal,
        defaults to ``max(times)*fs``

    kind : str
        Interpolation mode for the frequency and amplitude values.
        See: ``scipy.interpolate.interp1d`` for valid settings.

    Returns
    -------
    output : np.ndarray
        synthesized version of the pitch contour
    '''

    fs = float(fs)

    if length is None:
        length = int(times.max() * fs)

    # Squash the negative frequencies.
    # wave(0) = 0, so clipping here will un-voice the corresponding instants
    frequencies = np.maximum(frequencies, 0.0)

    # Build a frequency interpolator
    f_interp = interp1d(times * fs, 2 * np.pi * frequencies / fs, kind=kind,
                        fill_value=0.0, bounds_error=False, copy=False)

    # Estimate frequency at sample points
    f_est = f_interp(np.arange(length))

    if amplitudes is None:
        a_est = np.ones((length, ))
    else:
        # build an amplitude interpolator
        a_interp = interp1d(
            times * fs, amplitudes, kind=kind,
            fill_value=0.0, bounds_error=False, copy=False)
        a_est = a_interp(np.arange(length))

    # Sonify the waveform
    return a_est * function(np.cumsum(f_est))


def chroma(chromagram, times, fs, **kwargs):
    """Reverse synthesis of a chromagram (semitone matrix)

    Parameters
    ----------
    chromagram : np.ndarray, shape=(12, times.shape[0])
        Chromagram matrix, where each row represents a semitone [C->Bb]
        i.e., ``chromagram[3, j]`` is the magnitude of D# from ``times[j]`` to
        ``times[j + 1]``
    times: np.ndarray, shape=(len(chord_labels),) or (len(chord_labels), 2)
        Either the start time of each column in the chromagram,
        or the time interval corresponding to each column.
    fs : int
        Sampling rate to synthesize audio data at
    kwargs
        Additional keyword arguments to pass to
        :func:`mir_eval.sonify.time_frequency`

    Returns
    -------
    output : np.ndarray
        Synthesized chromagram

    """
    # We'll just use time_frequency with a Shepard tone-gram
    # To create the Shepard tone-gram, we copy the chromagram across 7 octaves
    n_octaves = 7
    # starting from C2
    base_note = 24
    # and weight each octave by a normal distribution
    # The normal distribution has mean 72 (one octave above middle C)
    # and std 6 (one half octave)
    mean = 72
    std = 6
    notes = np.arange(12*n_octaves) + base_note
    shepard_weight = np.exp(-(notes - mean)**2./(2.*std**2.))
    # Copy the chromagram matrix vertically n_octaves times
    gram = np.tile(chromagram.T, n_octaves).T
    # This fixes issues if the supplied chromagram is int type
    gram = gram.astype(float)
    # Apply Sheppard weighting
    gram *= shepard_weight.reshape(-1, 1)
    # Compute frequencies
    frequencies = 440.0*(2.0**((notes - 69)/12.0))
    return time_frequency(gram, frequencies, times, fs, **kwargs)


def chords(chord_labels, intervals, fs, **kwargs):
    """Synthesizes chord labels

    Parameters
    ----------
    chord_labels : list of str
        List of chord label strings.
    intervals : np.ndarray, shape=(len(chord_labels), 2)
        Start and end times of each chord label
    fs : int
        Sampling rate to synthesize at
    kwargs
        Additional keyword arguments to pass to
        :func:`mir_eval.sonify.time_frequency`

    Returns
    -------
    output : np.ndarray
        Synthesized chord labels

    """
    util.validate_intervals(intervals)

    # Convert from labels to chroma
    roots, interval_bitmaps, _ = chord.encode_many(chord_labels)
    chromagram = np.array([np.roll(interval_bitmap, root)
                           for (interval_bitmap, root)
                           in zip(interval_bitmaps, roots)]).T

    return chroma(chromagram, intervals, fs, **kwargs)
